
import abc, math, gc

import numpy as np, itertools, functools

from McUtils.Zachary import DensePolynomial, TensorDerivativeConverter
from McUtils.Scaffolding import Logger
from McUtils.Parallelizers import Parallelizer
import McUtils.Numputils as nput

__all__ = [
    "DGBEvaluator",
    "DGBKineticEnergyEvaluator",
    "DGBCartesianEvaluator",
    "DGBWatsonEvaluator",
    "DGBPotentialEnergyEvaluator",
    "DGBPairwisePotentialEvaluator",
    "DGBCartesianPairwiseEvaluator",
    "DGBWatsonPairwiseEvaluator"
]


class OverlapGaussianData:
    def __init__(self,
                 input_data,
                 product_data
                 ):
        self._input_data = input_data
        self._product_data = product_data
    @property
    def npts(self):
        return self.initial_centers.shape[0]
    @property
    def ndim(self):
        return self.initial_centers.shape[1]
    @property
    def centers(self):
        return self._product_data['centers']
    @property
    def alphas(self):
        return self._product_data['alphas']
    @property
    def rotations(self):
        return self._product_data['rotations']
    @property
    def inverse_rotations(self):
        return self._product_data['inverse_rotations']
    @property
    def precision_matrices(self):
        return self._product_data['covariances']
    @property
    def covariance_matrices(self):
        return self._product_data['inverse']
    @property
    def shift_matrices(self):
        return self._product_data['sum_inverse']
    @property
    def phases_diff(self):
        return self._product_data['phases_diff']
    @property
    def phases_sum(self):
        return self._product_data['phases_sum']
    @property
    def momenta_sum(self):
        return self._product_data['diagonal_momentum_sum']
    @property
    def momenta_diff(self):
        return self._product_data['diagonal_momentum_diff']
    @property
    def rho_sum(self):
        return self._product_data['rho_sum']
    @property
    def rho_diff(self):
        return self._product_data['rho_diff']
    @property
    def correlation_factors_sum(self):
        return self.cos_correlation_sum, self.sin_correlation_sum
    @property
    def correlation_factors_diff(self):
        return self.cos_correlation_diff, self.sin_correlation_diff
    @property
    def cos_correlation_diff(self):
        return self._product_data['cos_correlation_diff']
    @property
    def sin_correlation_diff(self):
        return self._product_data['sin_correlation_diff']
    @property
    def cos_correlation_sum(self):
        return self._product_data['cos_correlation_sum']
    @property
    def sin_correlation_sum(self):
        return self._product_data['sin_correlation_sum']
    @property
    def decay_factor_diff(self):
        return self._product_data['phase_diff_decay']
    @property
    def decay_factor_sum(self):
        return self._product_data['phase_sum_decay']
    @property
    def center_difference(self):
        return self._product_data['center_diffs']
    @property
    def delta_position(self):
        return self._product_data['center_delta']
    @property
    def delta_phase_sum(self):
        return self._product_data['phase_delta_sum']
    @property
    def delta_phase_diff(self):
        return self._product_data['phase_delta_diff']
    @property
    def indices(self):
        return self.row_indices, self.col_indices
    @property
    def row_indices(self):
        return self._product_data['row_inds']
    @property
    def col_indices(self):
        return self._product_data['col_inds']
    @property
    def initial_centers(self):
        return self._input_data['init_centers']
    @property
    def initial_alphas(self):
        return self._input_data['init_alphas']
    @property
    def initial_precision_matrices(self):
        return self._input_data['init_covariances']
    @property
    def initial_momenta(self):
        return self._input_data['initial_momenta']
    @property
    def initial_phases(self):
        return self._input_data['initial_phases']

    @classmethod
    def from_gaussian_parameters(cls,
                                 centers, alphas, transformations,
                                 momenta,
                                 *,
                                 chunk_size=None,  # int(1e6),
                                 rows_cols=None,
                                 logger=None,
                                 parallelizer=None
                                 ):
        return cls(*cls.get_overlap_data(
            centers, alphas, transformations,
            momenta,
            chunk_size=chunk_size,  # int(1e6),
            rows_cols=rows_cols,
            logger=logger,
            parallelizer=parallelizer
        ))

    @classmethod
    def get_overlap_data(clget_s,
                          centers, alphas, transformations,
                          aligned_momenta,
                          *,
                          chunk_size=None,  # int(1e6),
                          rows_cols=None,
                          logger=None,
                          parallelizer=None
                          ):
        parallelizer = Parallelizer.lookup(parallelizer)
        logger = Logger.lookup(logger)

        if rows_cols is None:
            rows, cols = np.triu_indices(len(alphas))
        else:
            rows, cols = rows_cols

        # a bit inefficient if all tfs are identity, but generality is good
        sigs = DGBEvaluator.get_inverse_covariances(alphas, transformations)
        covs = DGBEvaluator.get_covariances(alphas, transformations)
        momenta = DGBEvaluator.get_phase_vectors(aligned_momenta, transformations)
        new_sigs = sigs[rows] + sigs[cols]

        if chunk_size is not None:
            num_segments = new_sigs.shape[0] // chunk_size + 1
            chunks = np.array_split(new_sigs, num_segments)
            row_chunks = np.array_split(rows, num_segments)
            col_chunks = np.array_split(cols, num_segments)
        else:
            chunks = [new_sigs]
            row_chunks = [rows]
            col_chunks = [cols]

        logger.log_print("getting {nT} overlap Gaussians over {nC} chunks", nT=len(rows), nC=len(chunks))
        chunk_data = []
        for news, r, c in zip(chunks, row_chunks, col_chunks):
            new_alphas, new_rots = np.linalg.eigh(news)  # eigenvalues of inverse tensor...
            new_rots_inv = new_rots.transpose(0, 2, 1)

            # I _could_ construct the inverse from the alphas and rotations
            # but I think it makes more sense to use a potentially more stable
            # inverse here...also it is apparently no slower...
            new_inv = np.linalg.inv(news)
            # new_inv = new_rots @ nput.vec_tensordiag(1 / new_alphas) @ new_rots.transpose(0, 2, 1)

            new_centers = new_inv @ (
                    sigs[r] @ centers[r][:, :, np.newaxis]
                    + sigs[c] @ centers[c][:, :, np.newaxis]
            )
            new_centers = new_centers.reshape(centers[c].shape)
            new_alphas = new_alphas / 2
            sum_sigs = sigs[r] @ new_inv @ sigs[c]

            # TODO: turn this into a proper object...
            chunk_data.append({
                'centers': new_centers,
                'alphas': new_alphas,
                'covariances': new_sigs,
                'inverse': new_inv,
                'sum_inverse': sum_sigs,
                'rotations': new_rots,
                'inverse_rotations': new_rots_inv
            })


        product_data = {
            k: None if chunk_data[0][k] is None else np.concatenate([d[k] for d in chunk_data])
            for k in chunk_data[0].keys()
        }

        new_cov = product_data['inverse']
        new_si = product_data['sum_inverse']
        if momenta is not None:
            new_centers = product_data['centers']
            new_rots = product_data['rotations']
            new_alphas = product_data['alphas']

            # expressed in terms of the global frame
            phases_sum = momenta[rows] + momenta[cols]
            phases_diff = momenta[rows] - momenta[cols]
            rho_sum = np.reshape(new_cov @ phases_sum[:, :, np.newaxis], phases_sum.shape)
            rho_diff = np.reshape(new_cov @ phases_diff[:, :, np.newaxis], phases_sum.shape)
            # dot momentum sum or difference into displacement from origin
            corr_sum = np.reshape(new_centers[:, np.newaxis, :] @ phases_sum[:, :, np.newaxis], -1)
            corr_diff = np.reshape(new_centers[:, np.newaxis, :] @ phases_diff[:, :, np.newaxis], -1)
            # compute the J+ terms we use for all derivations
            scj = covs @ momenta[:, :, np.newaxis]
            delta_j_sum = np.reshape(new_si @ (scj[rows] - scj[cols]), phases_sum.shape)
            delta_j_diff = np.reshape(new_si @ (scj[rows] + scj[cols]), phases_diff.shape)

            # now express these in terms of the rotated frame

            mom_sum = DGBEvaluator.get_momentum_vectors(phases_sum, (new_rots, new_rots.transpose(0, 2, 1)))
            mom_diff = DGBEvaluator.get_momentum_vectors(phases_diff, (new_rots, new_rots.transpose(0, 2, 1)))

            decay_sum = np.sum(mom_sum ** 2 / (4 * new_alphas), axis=1)
            decay_diff = np.sum(mom_diff ** 2 / (4 * new_alphas), axis=1)

            sum_prefac = np.exp(-decay_sum)  # * np.cos(phase_sum)
            diff_prefac = np.exp(-decay_diff)  # * np.cos(phase_diff)
        else:
            mom_sum = None
            mom_diff = None
            phases_sum = None
            phases_diff = None
            rho_sum = None
            rho_diff = None
            corr_sum = None
            corr_diff = None
            sum_prefac = None
            diff_prefac = None

            delta_j_sum = None
            delta_j_diff = None

        row_gammas = new_cov @ sigs[rows]
        col_gammas = new_cov @ sigs[cols]
        disps = centers[rows] - centers[cols]
        delta = np.reshape(new_si @ disps[:, :, np.newaxis], disps.shape)
        product_data.update({
            'row_gammas':row_gammas,
            'col_gammas':col_gammas,
            'gamma_diff':row_gammas - col_gammas,
            'center_diffs':disps,
            'center_delta':delta,
            'diagonal_momentum_sum': mom_sum,
            'diagonal_momentum_diff': mom_diff,
            'phases_sum': phases_sum,
            'phases_diff': phases_diff,
            'rho_sum': rho_sum,
            'rho_diff': rho_diff,
            'phase_sum_corr': corr_sum,
            'phase_diff_corr': corr_diff,
            'phase_sum_decay': sum_prefac,
            'phase_diff_decay': diff_prefac,
            'phase_delta_sum':delta_j_sum,
            'phase_delta_diff':delta_j_diff,
            'row_inds': rows,
            'col_inds': cols,
        })
        if mom_sum is not None:
            product_data.update({
                'cos_correlation_sum': np.cos(corr_sum),
                'cos_correlation_diff': np.cos(corr_diff),
                'sin_correlation_sum': np.sin(corr_sum),
                'sin_correlation_diff': np.sin(corr_diff)
            })
        input_data = {
            'init_centers': centers,
            'init_alphas': alphas,
            'init_covariances': sigs,
            'initial_phases': momenta,
            'initial_momenta': aligned_momenta
        }

        return input_data, product_data

    def take_subselection(self, positions):
        # we need to find the positions in the old stuff where we
        rows, cols = self.indices

        mask_1, _, _ = nput.contained(rows, positions)
        mask_2, _, _ = nput.contained(cols, positions)
        take_pos = np.where(np.logical_and(mask_1 > 0, mask_2 > 0))[0]
        new_prod = {
            k: v[take_pos] if v is not None else v
            for k, v in self._product_data.items()
        }
        rows, cols = np.triu_indices(len(positions))
        new_prod['row_inds'] = rows
        new_prod['col_inds'] = cols

        new_input = {
            k: v[positions] if v is not None else v
            for k, v in self._input_data.items()
        }

        return type(self)(new_input, new_prod)
class DGBEvaluator:
    """
    An object that supports evaluating matrix elements in a distributed Gaussian basis.
    Provides support for integrating a function via quadrature or as an expansion in a polynomial tensors
    """

    @classmethod
    def get_inverse_covariances(cls, alphas, transformations):
        """
        Transforms the alphas into proper inverse covariance matrices.
        Chosen so that in the case that the transformations, Q, diagonalize S we can write
            QT S Q = A

        :return:
        :rtype:
        """

        n = alphas.shape[-1]
        npts = len(alphas)
        covs = np.zeros((npts, n, n))
        diag_inds = (slice(None, None, None),) + np.diag_indices(n)
        covs[diag_inds] = 2 * alphas

        if transformations is not None:
            tfs, inv = transformations
            covs = tfs @ covs @ tfs.transpose((0, 2, 1))

        return covs

    @classmethod
    def get_covariances(cls, alphas, transformations):
        """
        Transforms the alphas into proper inverse covariance matrices.
        Chosen so that in the case that the transformations, Q, diagonalize S we can write
            QT S Q = A

        :return:
        :rtype:
        """

        n = alphas.shape[-1]
        npts = len(alphas)
        covs = np.zeros((npts, n, n))
        diag_inds = (slice(None, None, None),) + np.diag_indices(n)
        covs[diag_inds] = 1/(2 * alphas)

        if transformations is not None:
            tfs, inv = transformations
            covs = tfs @ covs @ tfs.transpose((0, 2, 1))

        return covs

    @classmethod
    def get_momentum_vectors(cls, phases, transformations):
        """
        Transforms the momenta so that they're aligned along the Gaussian axes

        :return:
        :rtype:
        """

        if phases is not None and transformations is not None:
            tfs, inv = transformations
            momenta = np.reshape(
                phases[:, np.newaxis, :] @ inv.transpose((0, 2, 1)),
                phases.shape
            )
        else:
            momenta = phases

        return momenta

    @classmethod
    def get_phase_vectors(cls, momenta, transformations):
        """
        Transforms the momenta so that they're aligned along the Gaussian axes

        :return:
        :rtype:
        """

        if momenta is not None and transformations is not None:
            tfs, inv = transformations
            phases = np.reshape(
                momenta[:, np.newaxis, :] @ tfs.transpose((0, 2, 1)),
                momenta.shape
            )
        else:
            phases = momenta

        return phases

    @classmethod
    def get_overlap_gaussians(cls,
                              centers, alphas, transformations,
                              momenta,
                              *,
                              chunk_size=None,#int(1e6),
                              rows_cols=None,
                              logger=None,
                              parallelizer=None
                              ) -> 'OverlapGaussianData':
        return OverlapGaussianData.from_gaussian_parameters(
            centers, alphas, transformations,
            momenta,
            chunk_size=chunk_size,  # int(1e6),
            rows_cols=rows_cols,
            logger=logger,
            parallelizer=parallelizer
        )

    @classmethod
    def poch(cls, n, m): #pochammer/generalized gamma
        nums = np.arange(n-2*m+1 if m < n/3 else m + 1, n+1)
        dens = np.arange(1, m+1 if m < n/3 else n-2*m+1)
        if len(dens) < len(nums): # pad on left so we can have most stable eval
            dens = np.concatenate([
                np.ones(len(nums) - len(dens)),
                dens
            ])
        elif len(nums) < len(dens):
            nums = np.concatenate([
                np.ones(len(dens) - len(nums)),
                nums
            ])
        return np.prod(nums/dens)
    @classmethod
    def polyint_1D(cls, centers, alphas, n):
        if n == 0:
            return np.ones(centers.shape[:2])
        c = np.sqrt(alphas) * centers
        term = sum(
            (cls.poch(n, l) * (1/2**(2*l-n) if 2*l > n else 2**(n-2*l)))*(c)**(n-2*l)
            for l in range(0, int(np.floor(n/2)) + 1)
        )
        return term

    @classmethod
    def momentum_coeffient(cls, k, n):
        m = k//2
        s = (-1) ** k
        fac = (-1)**(n) * np.prod(
            2*np.arange(n+1, m+1, 2) - s
        )
        bin = math.comb(m, n)
        return bin * fac
    @classmethod
    def momentum_integral(cls, p, a, k):
        var = (p**2)/(2*a) # the rest of the sqrt(a) term is included elsewhere
        expansion = sum(
            cls.momentum_coeffient(k, n) * (var**n)
            for n in range(0, int(np.ceil(k/2)) + 1)
        ) / (2*a)**np.ceil(k/2)
        if k%2 == 1:
            expansion = p * expansion
        return expansion
    @classmethod
    def simple_poly_int(cls, n):
        return np.prod(np.arange(1, n, 2)) / 2**(n/2) # double factorial/gamma/whatever
    @classmethod
    def tensor_expansion_integrate(cls,
                                   npts, derivs, overlap_data:'OverlapGaussianData',
                                   expansion_type='multicenter',
                                   logger=None, reweight=True
                                   ):
        """
        provides an integral from a polynomial expansion with derivs as an expansion in tensors

        :param npts:
        :param derivs:
        :param centers:
        :param alphas:
        :param inds:
        :param rot_data:
        :param expansion_type:
        :param logger:
        :return:
        """

        centers = overlap_data.centers
        ndim = centers.shape[-1]
        rotations = overlap_data.rotations
        if expansion_type == 'taylor':
            centers = np.reshape(rotations @ centers[:, :, np.newaxis], centers.shape)
        alphas = overlap_data.alphas
        m_diff = overlap_data.momenta_diff # rotated along the Gaussian axes like the alphas
        m_sum = overlap_data.momenta_sum

        # if self.
        fdim = derivs[0].ndim - 1 # to simplify dipole functions
        fshape = derivs[0].shape[1:]

        new_derivs = []
        # rotations = rotations[:, :, :, np.newaxis] # to test shapes
        for n,d in enumerate(derivs):
            for _ in range(n):
                d = nput.vec_tensordot(
                    d, rotations,
                    axes=[fdim+1, 1],
                    shared=1
                )
            new_derivs.append(d)
        derivs = new_derivs

        if logger is not None:
            logger.log_print("adding up all derivative contributions...")

        row_inds = overlap_data.row_indices
        col_inds = overlap_data.col_indices

        pot = np.zeros((npts, npts) + fshape)
        caches = [{} for _ in range(ndim)]
        # mom_sum_pot = np.zeros((npts, npts) + fshape) # for momentum sum
        ms_caches = [{} for _ in range(ndim)]
        for nd,d in enumerate(derivs): # add up all independent integral contribs...
            # iterate over upper triangle coordinates (we'll add bottom contrib by symmetry)
            inds = itertools.combinations_with_replacement(range(ndim), r=nd) if nd > 0 else [()]
            for idx in inds:
                count_map = {k: v for k, v in zip(*np.unique(idx, return_counts=True))}
                if (
                        expansion_type != 'taylor' and m_sum is None
                        and any(n%2 !=0 for n in count_map.values())
                ):
                    continue # odd contribs vanish

                contrib = 1
                ms_contrib = 1
                for k in range(ndim): # do each dimension of integral independently
                    n = count_map.get(k, 0)
                    if n not in caches[k]:
                        if expansion_type == 'taylor':
                            raise NotImplementedError("Taylor series in rotated basis not implemented yet")

                        if m_sum is not None:
                            caches[k][n] = cls.momentum_integral(
                                m_diff[..., k],
                                alphas[..., k],
                                n
                            )
                            ms_caches[k][n] = cls.momentum_integral(
                                m_sum[..., k],
                                alphas[..., k],
                                n
                            )
                        else:
                            caches[k][n] = (
                               cls.simple_poly_int(n)
                                 if expansion_type != 'taylor' else
                               cls.polyint_1D(centers[..., k], alphas[..., k], n)
                            ) / alphas[..., k]**np.ceil(n/2)
                    base_contrib = caches[k][n]
                    for _ in range(fdim):
                        base_contrib = np.expand_dims(base_contrib, -1)
                    if isinstance(contrib, int) and fdim > 0:
                        base_contrib = np.broadcast_to(base_contrib,  alphas.shape[:-1] + fshape)
                    contrib *= base_contrib

                    if m_sum is not None:
                        base_contrib = ms_caches[k][n]
                        for _ in range(fdim):
                            base_contrib = np.expand_dims(base_contrib, -1)
                        if isinstance(ms_contrib, int) and fdim > 0:
                            base_contrib = np.broadcast_to(base_contrib, alphas.shape[:-1] + fshape)
                        ms_contrib *= base_contrib

                dcont = d[(slice(None, None, None),)*(fdim+1) + idx] if len(idx) > 0 else d
                if reweight:
                    # compute multinomial coefficient for weighting purposes
                    _, counts = np.unique(idx, return_counts=True)
                    multicoeff = 1
                    for x in counts: multicoeff *= math.factorial(x)
                    multicoeff = multicoeff / math.factorial(len(idx))
                    scaling = multicoeff / math.factorial(nd)
                    contrib *= scaling * dcont
                    if m_sum is not None:
                        ms_contrib *= scaling * dcont
                else:
                    contrib *= dcont
                    if m_sum is not None:
                        ms_contrib *= dcont

                if m_sum is not None:
                    sum_prefac = overlap_data.decay_factor_sum
                    diff_prefac = overlap_data.decay_factor_diff

                    contrib *= np.expand_dims(
                        overlap_data.cos_correlation_diff if nd%2 == 0 else -overlap_data.sin_correlation_diff,
                        [-x for x in range(1, fdim+1)]
                    )
                    ms_contrib *= np.expand_dims(
                        overlap_data.cos_correlation_sum if nd%2 == 0 else -overlap_data.sin_correlation_sum,
                        [-x for x in range(1, fdim+1)]
                    )

                    if ms_contrib.ndim > 1:
                        sum_prefac = sum_prefac[:, np.newaxis]
                        diff_prefac = diff_prefac[:, np.newaxis]

                    ms_contrib *= sum_prefac
                    contrib *= diff_prefac
                    # print("dd>", contrib[:5])

                    contrib = contrib + ms_contrib

                pot[row_inds, col_inds] += contrib

        pot[col_inds, row_inds] = pot[row_inds, col_inds]
        # mom_sum_pot[col_inds, row_inds] = mom_sum_pot[row_inds, col_inds]
        #
        # if m_sum is not None:
        #     return pot, mom_sum_pot
        # else:
        return pot

    @classmethod
    def quad_weight_eval(cls, function, d_chunk, w_chunk, ndim, centers, squa):
        d = d_chunk[np.newaxis] / squa[:, np.newaxis]
        c = centers[:, np.newaxis, :] + d
        fv = function(c.reshape(-1, ndim))
        if isinstance(fv, list):
            chunk_vals = []
            fshape = fv[0].shape[1:]
            w_chunk = np.expand_dims(w_chunk, [0] + [-x for x in range(1, len(fshape) + 1)])
            for mv in fv:
                mv = mv.reshape(c.shape[:2] + fshape)
                chunk_vals.append(np.sum(w_chunk * mv, axis=1))
            return chunk_vals
        else:
            fshape = fv.shape[1:]
            fv = fv.reshape(c.shape[:2] + fshape)
            w_chunk = np.expand_dims(w_chunk, [0] + [-x for x in range(1, len(fshape) + 1)])
            chunk_val = np.sum(w_chunk * fv, axis=1)
            return chunk_val
    @classmethod
    def quad_nd(cls,
                centers, alphas, function,
                flatten=False,
                degree=3, chunk_size=int(1e6),
                normalize=True
                ):
        """
        N-dimensional quadrature

        :param centers:
        :param alphas:
        :param function:
        :param degree:
        :return:
        """

        centers = np.asanyarray(centers)
        if centers.ndim == 1:
            centers = centers[np.newaxis]
            alphas = [alphas]
        alphas = np.asanyarray(alphas)

        # Quadrature point displacements and weights (thanks NumPy!)
        disps, weights = np.polynomial.hermite.hermgauss(degree)

        ndim = centers.shape[-1]
        indices = np.moveaxis(np.array(
                np.meshgrid(*([np.arange(0, degree, dtype=int)] * ndim))
            ), 0, -1).reshape(-1, ndim)
        disps = disps[indices]

        w = np.prod(weights[indices], axis=-1)
        n_disps = alphas.shape[0] * disps.shape[0] # how many structures total
        num_segments = n_disps // chunk_size + 1
        disp_chunks = np.array_split(disps, num_segments)
        w_chunks = np.array_split(w, num_segments)

        val = None
        squa = np.sqrt(alphas)
        for d_chunk, w_chunk in zip(disp_chunks, w_chunks):
            chunk_val = cls.quad_weight_eval(function, d_chunk, w_chunk, ndim, centers, squa)
            if val is None:
                val = chunk_val
            else:
                if isinstance(val, list):
                    for i in range(len(val)):
                        val[i] += chunk_val[i]
                else:
                    val += chunk_val

        if normalize:
            normalization = 1 / np.prod(squa, axis=-1)
            if isinstance(val, list):
                for i in range(len(val)):
                    val[i] *= normalization
            else:
                val *= normalization
        return val

    @classmethod
    def _wrap_rotated_function(cls, func, rotations, inverse, momenta):

        # momenta = momenta@rotations
        def eval(coords, deriv_order=None, inverse=inverse, rotations=rotations):
            # coords are expressed in the rotated space, we need to transform back out
            quad_disps = coords.shape[0] // inverse.shape[0]
            inverse = np.broadcast_to(
                inverse[:, np.newaxis, :, :],
                (inverse.shape[0], quad_disps) + inverse.shape[1:]
            ).reshape((coords.shape[0],) + inverse.shape[1:])
            unrot_coords = np.reshape(inverse @ coords[:, :, np.newaxis], coords.shape)
            fdat = func(unrot_coords, deriv_order=deriv_order)
            if deriv_order is not None:
                rotations = np.broadcast_to(
                    rotations[:, np.newaxis, :, :],
                    (rotations.shape[0], quad_disps) + rotations.shape[1:]
                ).reshape((coords.shape[0],) + rotations.shape[1:])

                f_vals = fdat[0]
                f_derivs = []
                for n, d in enumerate(fdat[1:]):
                    # now we want to rexpress the derivatives in the rotated space (I think)
                    for _ in range(n):
                        d = nput.vec_tensordot(
                            d, rotations,
                            axes=[1, 1],
                            shared=1
                        )
                    f_derivs.append(d)
                fdat = [f_vals] + f_derivs
            if momenta is not None:
                # new_dat = []
                fvals = fdat if isinstance(fdat, np.ndarray) else fdat[0]
                cos_cor = 0
                for p in momenta:
                    p = np.broadcast_to(
                        p[:, np.newaxis, :],
                        (p.shape[0], quad_disps) + p.shape[1:]
                    ).reshape((unrot_coords.shape[0],) + p.shape[1:])
                    correlation = np.reshape(
                        np.matmul(
                            unrot_coords[:, np.newaxis, :],
                            p[:, :, np.newaxis]
                        ),
                        -1
                    )
                    cos_cor += np.cos(correlation)
                new_dat = cos_cor * fvals
                if isinstance(fdat, np.ndarray):
                    fdat = new_dat
                else:
                    fdat[0] = new_dat
            return fdat

        return eval

    @classmethod
    def rotated_gaussian_quadrature(cls,
                                    function,
                                    alphas,
                                    centers,
                                    rotations,
                                    inverse,
                                    momenta,
                                    normalize=True,
                                    degree=2,
                                    ):
        f = cls._wrap_rotated_function(function, rotations, inverse, momenta)
        centers = np.reshape(rotations @ centers[:, :, np.newaxis], centers.shape)

        vals = cls.quad_nd(centers, alphas, f, degree=degree, normalize=normalize)

        normalization = 1 / (np.sqrt(np.pi)) ** centers.shape[-1]  # part of the prefactor...
        vals *= normalization
            # diff_vals, sum_vals = vals
            # diff_phase, sum_phase = momenta
            # diff_corr = np.reshape(
            #     np.matmul(centers[:, np.newaxis, :], diff_phase[:, :, np.newaxis]),
            #     -1
            # )
            # sum_corr = np.reshape(
            #     np.matmul(centers[:, np.newaxis, :], sum_phase[:, :, np.newaxis]),
            #     -1
            # )
            # print(sum_vals)
            # vals = [
            #     diff_vals * np.cos(diff_corr),
            #     sum_vals * np.cos(sum_corr)
            # ]


        return vals

    @classmethod
    def quad_integrate(cls, function, overlap_data:'OverlapGaussianData', degree=2, logger=None):
        """
        Integrate potential over all pairs of Gaussians at once

        :param degree:
        :type degree:
        :return:
        :rtype:
        """

        use_momenta = overlap_data.momenta_diff is not None
        vals = cls.rotated_gaussian_quadrature(
            function,
            overlap_data.alphas,
            overlap_data.centers,
            overlap_data.inverse_rotations,
            overlap_data.rotations,
            [overlap_data.phases_diff, overlap_data.phases_sum]
                if use_momenta else None,
            degree=degree,
            normalize=False
        )

        npts = overlap_data.npts
        rows, cols = np.triu_indices(npts)

        pots = np.zeros((npts, npts) + vals.shape[1:])
        pots[rows, cols] = vals
        pots[cols, rows] = vals

        return pots

    @classmethod
    def _rot_base_S_components(cls, overlap_data:'OverlapGaussianData'):
        row_inds = overlap_data.row_indices
        col_inds = overlap_data.col_indices
        alphas = overlap_data.initial_alphas
        rotas = overlap_data.alphas

        ndim = overlap_data.ndim
        # we prefactor out the 2**ndim

        det_rat = np.prod(
            np.sort(alphas[row_inds]/rotas, axis=1)
            * np.flip(np.sort(alphas[col_inds]/rotas, axis=1), axis=1),
            axis=1
        )

        thing1 = np.prod(alphas[row_inds]/rotas, axis=1)
        n = len(alphas)
        rat_mat = np.zeros((n, n))
        rat_mat[row_inds, col_inds] = thing1

        disps = overlap_data.center_difference
        C = disps[:, np.newaxis, :] @ overlap_data.shift_matrices @ disps[:, :, np.newaxis]
        C = C.reshape(disps.shape[0])

        return row_inds, col_inds, ndim, det_rat, C

    @classmethod
    def _rot_base_S(cls, overlap_data:'OverlapGaussianData', logger=None, return_prefactor=False):

        logger.log_print('evaluating {nT} overlaps', nT=len(overlap_data.row_indices))
        n = overlap_data.npts
        S = np.eye(n)
        row_inds, col_inds, ndim, det_rat, C = cls._rot_base_S_components(overlap_data)

        scaled_rats = ( 2**(ndim/2) ) * ( (det_rat) ** (1 / 4) )
        exp_terms = np.exp(-C / 2)
        full_terms = scaled_rats * exp_terms

        # include phase factors
        if overlap_data.initial_momenta is not None:
            j_vecs = overlap_data.initial_phases

            phases = np.reshape(overlap_data.initial_centers[:, np.newaxis, :] @ j_vecs[:, :, np.newaxis], -1)
            decay_norm = np.sum(overlap_data.initial_momenta**2 / (2*overlap_data.initial_alphas), axis=1)
            norm = np.sqrt((1 + np.cos(2*phases)*np.exp(-decay_norm)))
            norms = norm[row_inds] * norm[col_inds]

            full_terms = full_terms / norms

            prefac = np.eye(n)
            prefac[row_inds, col_inds] = full_terms
            prefac[col_inds, row_inds] = prefac[row_inds, col_inds]

            decay_contrib = (
                    overlap_data.decay_factor_diff * overlap_data.cos_correlation_diff
                    + overlap_data.decay_factor_sum * overlap_data.cos_correlation_sum
            )
            S[row_inds, col_inds] = decay_contrib
            S[col_inds, row_inds] = decay_contrib
            S = S * prefac

        else:
            S[row_inds, col_inds] = full_terms
            S[col_inds, row_inds] = full_terms
            prefac = S

        if return_prefactor:
            return prefac, S
        else:
            return S

    @classmethod
    def evaluate_overlap(cls, overlap_data:'OverlapGaussianData', logger=None, return_prefactor=False):
        return cls._rot_base_S(overlap_data, logger=logger, return_prefactor=return_prefactor)

class DGBKineticEnergyEvaluator(DGBEvaluator):
    """
    """

    @classmethod
    def _evaluate_polynomial_ke(cls, overlap_data:'OverlapGaussianData', terms, prefactors):
        # allows us to create a tensor of prefactors and a tensor of terms and induce
        # a polynomial by the action of the terms and then multiply that elementwise by
        # the prefactors to build the total polynomial expansion we can integrate
        # rows = overlap_data.row_indices
        cols = overlap_data.col_indices
        centers = overlap_data.centers
        # alphas = overlap_data.alphas
        init_sigs = overlap_data.initial_precision_matrices
        init_cents = overlap_data.initial_centers

        nz_pos = np.where(prefactors != 0)
        if len(nz_pos) == 0 or len(nz_pos[0]) == 0:
            # short circuit
            return np.zeros((init_cents.shape[0], init_cents.shape[0]))

        # ndim = centers.shape[-1]
        poly = cls._induced_polys(terms, init_sigs[cols], init_cents[cols], centers)
        pcs = poly.coeffs
        if not isinstance(pcs, np.ndarray):
            pcs.asarray()
        # force prefactors to be broadcastable
        scaled_coeffs = np.expand_dims(
            prefactors[nz_pos],
            list(range(1, 1 + poly.coordinate_dim))
        ) * pcs[nz_pos]
        poly = DensePolynomial(scaled_coeffs, stack_dim=1)
        # print(poly.coeffs[0])
        # poly = poly.shift(-centers[nz_pos[0]]).make_sparse_backed()
        # raise Exception(centers[0], alphas[0], init_sigs[0])
        tcoeffs = poly.coefficient_tensors
        (keys, subsel), sorting = nput.group_by(np.arange(len(nz_pos[0])), nz_pos[0])
        if len(keys) != len(centers):
            raise NotImplementedError("no contrib case not yet handled...")
        # we assume keys is sorted, so we just need to iterate over the subsels
        tensors = [
            [np.sum(x[sel], axis=0) for x in tcoeffs]
            for sel in subsel
        ]
        # nput.group_by(, nz_pos[0])
        # poly = poly.shift(-np.expand_dims(centers, list(range(1, len(terms)+1))))

        # # raise Exception(poly.coeffs.shape)
        # for pc, prefs in zip(poly.coeffs, prefactors):
        #     nz_pos = np.where(prefs != 0)
        #     if len(nz_pos) == 0:
        #         raise NotImplementedError("no contrib case not yet handled...")
        #     sub_poly = DensePolynomial(prefs[nz_pos] * pc[nz_pos], stack_dim=1)
        #     print(sub_poly.coefficient_tensors[0], sub_poly.coefficient_tensors[1])
        #     tc_sum = [np.sum(x, axis=0) for x in sub_poly.coefficient_tensors]
        #     tensors.append(tc_sum)

        # for ri, ci, pref, new_cent in zip(rows, cols, prefactors, centers): #TODO: vectorize most of this
        #     sig = init_sigs[ci]
        #     cent = init_cents[ci]
        #     polys = prefactors * cls._induced_polys(terms, ndim, sig, cent)
        #     expansion = None
        #     for p in polys.flat:
        #         p = p.shift(-new_cent)
        #         if expansion is None:
        #             expansion = p.coefficient_tensors
        #         else:
        #             expansion = [p1+p2 for p1, p2 in zip(expansion, p.coefficient_tensors)]
        #     tensors.append(expansion)

        tensors = [
            np.array([t[i] for t in tensors])
            for i in range(len(tensors[0]))
        ]
        # print(tensors[0][:3].flatten())
        # print(tensors[2][:3].flatten())

        return -1/2*cls.tensor_expansion_integrate(init_cents.shape[0], tensors, overlap_data, reweight=False)

    @classmethod
    def _induced_polys(cls, terms, sigs, initial_centers, final_centers):
        centering_shift = initial_centers - final_centers

        sig_polys = DensePolynomial.from_tensors(
            [np.zeros(sigs.shape[0]), np.zeros(sigs.shape[:2]), -sigs/2]
        ).shift(centering_shift)#.make_sparse_backed(1e-10)
        sig_grad = sig_polys.grad() # for applying p

        if 'q' in terms:
            monomial = DensePolynomial.from_tensors( # for applying q
                [np.zeros(sig_polys.coordinate_dim), np.eye(sig_polys.coordinate_dim)]
            )
            mcs = monomial.coeffs
            if not isinstance(mcs, np.ndarray):
                mcs = mcs.asarray()
            monomial = DensePolynomial(
                np.broadcast_to(mcs[np.newaxis], (sigs.shape[0],) + mcs.shape),
                stack_dim=2
            )
            monomial = monomial.shift(-final_centers[:, np.newaxis, :])#.make_sparse_backed(1e-10)

        else:
            monomial = None

        poly = DensePolynomial(
            np.ones((len(sigs),) + (1,) * sig_polys.coordinate_dim),
            stack_dim=1
        )#.make_sparse_backed(1e-10) # stack_dim = 1
        for t in terms:
            # _ = []
            # from Peeves.Timer import Timer
            # with Timer(tag=f"term {t}"):
            if t == 'q':

                mon_coeffs = monomial.coeffs
                if isinstance(mon_coeffs, np.ndarray):
                    mon_coeffs = np.expand_dims(mon_coeffs, list(range(1, poly.stack_dim)))
                else:
                    mon_coeffs = mon_coeffs.expand_dims(list(range(1, poly.stack_dim)))

                poly_coeffs= poly.coeffs
                if isinstance(poly_coeffs, np.ndarray):
                    poly_coeffs = np.expand_dims(poly_coeffs, poly.stack_dim)
                else:
                    poly_coeffs = poly_coeffs.expand_dims(poly.stack_dim)

                poly = DensePolynomial(
                    poly_coeffs,
                    stack_dim=poly.stack_dim + 1
                ) * DensePolynomial(
                    mon_coeffs,
                    stack_dim=poly.stack_dim + 1
                )
                # for i in range(ndim):
                #     monomial = DensePolynomial(np.zeros((2,) * ndim))
                #     if coeffs is None:
                #         new = monomial
                #     else: # coeffs is a NumPy array of objects
                #         new = np.apply_along_axis(
                #             lambda poly_slice:cls._apply_q(poly_slice, monomial),
                #             -1,
                #             coeffs
                #         )
                #     _.append(new)

                # print(poly.coeffs)
            elif t == 'p':
                # first term is just taking derivs of coeffs
                # basic idea is each time we hit a p we have
                # d/dq_i(K * e^(sigma*q*q)) = (d/dq_i(K) + K * ??? )e^(sigma*q*q)
                poly_1 = poly.grad()

                sigg_coeffs = sig_grad.coeffs
                if isinstance(sigg_coeffs, np.ndarray):
                    sigg_coeffs = np.expand_dims(sigg_coeffs, list(range(1, poly.stack_dim)))
                else:
                    sigg_coeffs = sigg_coeffs.expand_dims(list(range(1, poly.stack_dim)))

                poly_coeffs = poly.coeffs
                if isinstance(poly_coeffs, np.ndarray):
                    poly_coeffs = np.expand_dims(poly_coeffs, poly.stack_dim)
                else:
                    poly_coeffs = poly_coeffs.expand_dims(poly.stack_dim)

                pad_coeffs = DensePolynomial(
                    poly_coeffs,
                    stack_dim=poly.stack_dim + 1
                )
                pad_grad = DensePolynomial(
                    sigg_coeffs,
                    stack_dim=poly.stack_dim + 1
                )
                # raise Exception(pad_coeffs, pad_grad)
                poly_2 = pad_coeffs * pad_grad
                poly = poly_1 + poly_2
                # cfs = poly.coeffs
                    # raise Exception(cfs.shape)
            # for i in range(ndim):
            #     if coeffs is None:
            #         new = sig_polys.deriv(i)
            #     else: # coeffs is a NumPy array of objects
            #         new = np.apply_along_axis(
            #             lambda poly_slice:cls._apply_p(poly_slice, sig_polys, i), -1, coeffs
            #         )
            #     _.append(new)
            else:
                raise ValueError("don't understand term {t}".format(t=t))

            # cshape = () if coeffs is None else coeffs.shape
            # coeffs = np.full((ndim,) + cshape, None, dtype=object)
            # for i, p in enumerate(_):
            #     coeffs[i] = p
            # print(coeffs.shape, coeffs)

        return poly

    @abc.abstractmethod
    def evaluate_ke(self, overlap_data:'OverlapGaussianData', logger=None, **opts):
        ...

    @classmethod
    def evaluate_diagonal_rotated_momentum_contrib(self, overlap_data:'OverlapGaussianData', masses):
        # polynomial eval is too slow, need to fix in general but this is faster in the meantime
        si_cov = overlap_data.shift_matrices
        diag_inds = np.diag_indices_from(si_cov[0])
        delta_vec = overlap_data.delta_position
        shift_contrib = (delta_vec)**2
        diag_inds = (slice(None),) + diag_inds
        diag_contrib = si_cov[diag_inds]

        base_contrib = diag_contrib - shift_contrib

        j = overlap_data.initial_phases
        if j is not None:
            jp_diff = overlap_data.delta_phase_diff
            jp_sum = overlap_data.delta_phase_sum

            cos_sum = overlap_data.cos_correlation_sum
            cos_diff = overlap_data.cos_correlation_diff
            sin_sum = overlap_data.sin_correlation_sum
            sin_diff = overlap_data.sin_correlation_diff

            diff_contrib = 1 / 2 * np.dot(
                cos_diff[:, np.newaxis]*(base_contrib+jp_diff**2)
                + sin_diff[:, np.newaxis] * (2*jp_diff*delta_vec), # factored out the negative sign already
                1 / masses
            )
            sum_contrib = 1 / 2 * np.dot(
                cos_sum[:, np.newaxis] * (base_contrib+jp_sum**2)
                + sin_sum[:, np.newaxis] * (2*jp_sum*delta_vec), # factored out the negative sign already
                1 / masses
            )

            sum_prefac = overlap_data.decay_factor_sum
            diff_prefac = overlap_data.decay_factor_diff
            contrib = diff_prefac * diff_contrib + sum_prefac * sum_contrib

        else:
            contrib = 1 / 2 * np.dot(base_contrib, 1 / masses)

        n = overlap_data.npts
        rows, cols = overlap_data.indices
        ke = np.zeros((n, n))
        ke[rows, cols] = contrib
        ke[cols, rows] = contrib
        return ke

class DGBCartesianEvaluator(DGBKineticEnergyEvaluator):

    def __init__(self, masses):
        self.masses = masses

    def evaluate_ke(self, overlap_data:'OverlapGaussianData', logger=None):
        logger = Logger.lookup(logger)
        logger.log_print("evaluating Cartesian kinetic energy contribution")

        masses = self.masses
        return self.evaluate_diagonal_rotated_momentum_contrib(
            overlap_data,
            masses
        )

        # prefactors = np.broadcast_to(
        #     np.diag(1 / (masses))[np.newaxis],
        #     (overlap_data.centers.shape[0], len(masses), len(masses))
        # )
        # return self._evaluate_polynomial_ke(overlap_data, ['p', 'p'], prefactors)

class DGBWatsonEvaluator(DGBKineticEnergyEvaluator):
    """
    """
    def __init__(self, modes, coriolis_inertia_function):
        self.modes = modes
        self.ci_func = coriolis_inertia_function

    @classmethod
    def _embed_watson_modes(cls, watson_data, centers): # for putting cartesians in the watson frame?
        origin = watson_data['origin'].reshape(-1)
        tf = watson_data['matrix']
        new_cent_vecs = (centers-origin[np.newaxis]).reshape(centers.shape[0], -1, centers.shape[1])@tf.T[np.newaxis]
        return new_cent_vecs.reshape(-1, new_cent_vecs.shape[-1]) # basically a squeeze

    @staticmethod
    def annoying_coriolis_term(
            n, u, m, v,

            Xc, Dx, Sc, Sp,
            Gi, Gj, DG
    ):
        return (
                (Xc[:, m] * Xc[:, n] + Sc[:, n, m]) * (Dx[:, v] * Dx[:, u] - Sp[:, u, v])
                - (Gi[:, m, v]*Gj[:, n, u] + Gi[:, m, u]*Gj[:, n, u])
                - (Dx[:, v]*Xc[:, n]*DG[:, m, u] + Dx[:, v]*Xc[:, m]*DG[:, n, u])
        )
    @staticmethod
    def annoying_coriolis_momentum_term(
            n, u, m, v,

            Xc, r, Jp, Dx,
            Sc, Sp, DG
            # Gi, Gj, DG
    ):
        return (
            Jp[:, u]*Jp[:, v]*(Xc[:, m]*Xc[:, n] - r[:, m]*r[:, m] + Sc[:, n, m])
            -r[:, m]*r[:, n]*(Dx[:, v] * Dx[:, u] - Sp[:, u, v])
            -(Xc[:, n]*r[:, m] + Xc[:, m]*r[:, n])*(Dx[:, v]*Jp[:, u] + Dx[:, u]*Jp[:, v])
            +Jp[:, v]*r[:, n]*DG[:, m, u] + Jp[:, v]*r[:, m]*DG[:, n, u]
        )
    @staticmethod
    def annoying_imaginary_momentum_term(
            n, u, m, v,

            Xc, r, Jp, Dx,
            Sc, Sp, DG
            # Gi, Gj, DG
    ):
        return (
            -(Xc[:, n]*r[:, m] + Xc[:, m]*r[:, n])*(Dx[:, v] * Dx[:, u] - Sp[:, u, v] - Jp[:, u]*Jp[:, v])
            -(Xc[:, m]*Xc[:, n] - r[:, n]*r[:, m] + Sc[:, n, m])*(Dx[:, v]*Jp[:, u] + Dx[:, u]*Jp[:, v])
            +(Jp[:, v]*Xc[:, n]+Dx[:, v]*r[:, n])*DG[:, m, u]
            +(Jp[:, v]*Xc[:, m]+Dx[:, v]*r[:, m])*DG[:, n, u]
        )
    @classmethod
    def evaluate_coriolis_contrib(cls, coriolis_tensors, overlap_data:'OverlapGaussianData'):

        Sc = overlap_data.covariance_matrices
        Sp = overlap_data.shift_matrices
        SI = overlap_data.initial_precision_matrices
        Xc = overlap_data.centers

        rows, cols = overlap_data.indices
        SIi = SI[rows]
        SIj = SI[cols]

        Gi = Sc @ SIi
        Gj = Sc @ SIj
        DG = Gi - Gj

        Dx = overlap_data.center_difference
        Dx = np.reshape(Sp @ overlap_data.center_difference[:, :, np.newaxis], Dx.shape)

        ndim = Xc.shape[-1]

        term = lambda n, u, m, v: coriolis_tensors[:, n, u, m, v] * cls.annoying_coriolis_term(
            n, u, m, v,

            Xc, Dx, Sc, Sp,
            Gi, Gj, DG
        )

        contrib = np.zeros(Sc.shape[0])
        inds = itertools.product(*[range(ndim)]*4)
        for n,u,m,v in inds:
            if n != u and m != v:
                t = term(n, u, m, v)
                contrib += t

        if overlap_data.initial_momenta is not None:
            jp_sum = overlap_data.delta_phase_sum
            jp_diff = overlap_data.delta_phase_diff
            rho_sum = overlap_data.rho_sum
            rho_diff = overlap_data.rho_diff

            sum_real_term = lambda n, u, m, v: coriolis_tensors[:, n, u, m, v] * cls.annoying_coriolis_momentum_term(
                    n, u, m, v,

                    Xc, rho_sum, jp_sum, Dx,
                    Sc, Sp, DG
            )
            sum_imag_term = lambda n, u, m, v: coriolis_tensors[:, n, u, m, v] * cls.annoying_imaginary_momentum_term(
                n, u, m, v,

                Xc, rho_sum, jp_sum, Dx,
                Sc, Sp, DG
            )

            diff_real_term = lambda n, u, m, v: coriolis_tensors[:, n, u, m, v] * cls.annoying_coriolis_momentum_term(
                n, u, m, v,

                Xc, rho_diff, jp_diff, Dx,
                Sc, Sp, DG
            )
            diff_imag_term = lambda n, u, m, v: coriolis_tensors[:, n, u, m, v] * cls.annoying_imaginary_momentum_term(
                n, u, m, v,

                Xc, rho_diff, jp_diff, Dx,
                Sc, Sp, DG
            )

            sum_real_contrib = np.zeros(Sc.shape[0])
            sum_imag_contrib = np.zeros(Sc.shape[0])
            diff_real_contrib = np.zeros(Sc.shape[0])
            diff_imag_contrib = np.zeros(Sc.shape[0])
            inds = itertools.product(*[range(ndim)] * 4)
            for n, u, m, v in inds:
                if n != u and m != v:
                    real_cont = sum_real_term(n, u, m, v)
                    imag_cont = sum_imag_term(n, u, m, v)
                    # print("Sum I:", (n, u, m, v), 219475*imag_cont)
                    # print("Sum R:", (n, u, m, v), 219475*real_cont)
                    sum_real_contrib += real_cont
                    sum_imag_contrib += imag_cont
                    real_cont = diff_real_term(n, u, m, v)
                    imag_cont = diff_imag_term(n, u, m, v)
                    # print("Diff I:", (n, u, m, v), 219475*imag_cont)
                    # print("Diff R:", (n, u, m, v), 219475*real_cont)
                    diff_real_contrib += real_cont
                    diff_imag_contrib += imag_cont

            cos_sum, sin_sum = overlap_data.correlation_factors_sum
            cos_diff, sin_diff = overlap_data.correlation_factors_diff
            sum_prefac = overlap_data.decay_factor_sum
            diff_prefac = overlap_data.decay_factor_diff

            diff_contrib = (
                    cos_diff * (contrib + diff_real_contrib)
                    + sin_diff * diff_imag_contrib
            )
            sum_contrib = (
                    cos_sum * (contrib + sum_real_contrib)
                    + sin_sum * sum_imag_contrib
            )

            contrib = diff_prefac * diff_contrib + sum_prefac * sum_contrib

        npts = overlap_data.npts
        ke = np.zeros((npts, npts))
        ke[rows, cols] = -contrib
        ke[cols, rows] = -contrib

        return ke

    @classmethod
    def evaluate_watson_term(cls, B_e, overlap_data:'OverlapGaussianData'):
        B_e = -1 / 4 * B_e
        if overlap_data.initial_momenta is not None:
            phase_sum = overlap_data.cos_correlation_sum
            phase_diff = overlap_data.cos_correlation_diff

            sum_prefac = overlap_data.decay_factor_sum
            diff_prefac = overlap_data.decay_factor_diff

            diff_wat = np.cos(phase_diff)
            sum_wat = np.cos(phase_sum)
            watson_contrib = B_e * (diff_prefac * diff_wat + sum_prefac * sum_wat)
        else:
            watson_contrib = B_e

        n = overlap_data.npts
        watson = np.zeros((n, n))
        rows, cols = overlap_data.indices
        watson[rows, cols] = watson_contrib
        watson[cols, rows] = watson_contrib

        return watson

    def evaluate_ke(self, overlap_data: 'OverlapGaussianData',
                    logger=None,
                    include_diagonal_contribution=True,
                    include_coriolis_coupling=True,
                    include_watson_term=True
                    ):

        logger = Logger.lookup(logger)
        logger.log_print("evaluating Watson diagonal momentum contribution")
        if include_diagonal_contribution:
            base = self.evaluate_diagonal_rotated_momentum_contrib(
                overlap_data,
                np.ones(overlap_data.ndim)
            )
        else:
            base = 0

        if include_watson_term or include_coriolis_coupling:
            if include_coriolis_coupling:
                logger.log_print("evaluating coriolis contribution")
            B_e, coriolis = self.ci_func(overlap_data.centers)
            if include_coriolis_coupling:
                coriolis = self.evaluate_coriolis_contrib(coriolis, overlap_data)
            else:
                coriolis = 0
        else:
            coriolis = 0
            B_e = 0

        if include_watson_term:
            logger.log_print("evaluating Watson term contribution")
        if include_watson_term:
            watson = self.evaluate_watson_term(B_e, overlap_data)
        else:
            watson = 0

        ke = base + watson + coriolis

        return ke

class DGBPotentialEnergyEvaluator(DGBEvaluator):
    """
    An evaluator designed
    """

    def __init__(self,
                 potential_function,
                 integral_handler=None,
                 expansion_degree=None,
                 expansion_type=None,
                 quadrature_degree=None,
                 pairwise_functions=None,
                 logger=None
                 ):
        self.potential_function = potential_function
        self.handler = integral_handler
        self.expansion_degree = expansion_degree
        self.expansion_type = expansion_type
        self.quadrature_degree = quadrature_degree
        self.pairwise_handler = pairwise_functions

        self.logger = Logger.lookup(logger)

    def analytic_integrate(self):
        raise NotImplementedError("flooped up")
        centers = [np.array(np.meshgrid(x, x)).T for x in self.centers.T]
        alphas = np.array(np.meshgrid(self.alphas, self.alphas)).T
        # raise Exception(alphas.shape)
        return self.potential_function['analytic_integrals'](
            centers, # ...no
            alphas
        )

    @classmethod
    def expansion_integrate(cls,
                            function,
                            overlap_data:'OverlapGaussianData',
                            expansion_type,
                            expansion_degree=2,
                            pairwise_functions=None,
                            logger=None
                            ):

        if pairwise_functions is not None:
            logger.log_print("evaluating pairwise potential contributions...")
            pot_contribs, deriv_corrs = pairwise_functions.evaluate_pairwise_contrib(
                overlap_data,
                expansion_degree=None if expansion_degree < 0 else expansion_degree
            )
        else:
            pot_contribs = None
            deriv_corrs = None

        if not isinstance(logger, Logger):
            logger = Logger.lookup(logger)

        if expansion_degree > -1:

            deriv_order = expansion_degree # renamed for consistency
            if expansion_type == 'taylor':
                raise NotImplementedError('need to fix up taylor expansion')
                logger.log_print("expanding as a Taylor series about the minumum energy geometry...")
                assert self.ref is None #TODO: centers need a displacement
                zero = np.zeros((1, centers.shape[-1])) if self.ref is None else np.array([self.ref])
                derivs = function(zero, deriv_order=deriv_order)
                if isinstance(derivs, np.ndarray): # didn't get the full list so we do the less efficient route
                    derivs = (
                            [function(zero)] +
                            [function(zero, deriv_order=d) for d in range(1, deriv_order + 1)]
                    )
                derivs = [
                    np.broadcast_to(
                        d[np.newaxis],
                        alphas.shape + d.squeeze().shape
                    )
                    for d in derivs
                ]
            else:
                alphas = overlap_data.alphas
                centers = overlap_data.centers
                if overlap_data.momenta_sum is None:
                    deriv_order = deriv_order - (deriv_order % 2) # odd orders don't contribute so why evaluate the derivatives...
                # if self.mass_weighted:
                #     derivs = self.mass_weighted_eval(function, centers, self.masses, deriv_order=deriv_order)
                # else:
                derivs = function(centers, deriv_order=deriv_order)
                if isinstance(derivs, np.ndarray):  # didn't get the full list so we do the less efficient route'
                    derivs = [function(centers)] + [
                        function(centers, deriv_order=d) for d in range(1, deriv_order+1)
                    ]

            if deriv_corrs is not None:
                _ = []
                for i,(d1,d2) in enumerate(zip(derivs, deriv_corrs)):
                    _.append(d1 - d2)
                derivs = _

            npts = overlap_data.npts
            pot = cls.tensor_expansion_integrate(
                npts,
                derivs,
                overlap_data,
                reweight=True,
                expansion_type=expansion_type,
                logger=logger
            )

        else:
            npts = overlap_data.npts
            pot = np.zeros((npts, npts))

        # if prefactors is not None:
        #     row_inds, col_inds = np.triu_indices(npts)
        #     pref = np.zeros((npts, npts))
        #     pref[row_inds, col_inds] = prefactors
        #     pref[col_inds, row_inds] = prefactors
        #     pot *= pref

        if pot_contribs is not None: # implies no rotation
            row_inds, col_inds = np.triu_indices(npts)
            pot2 = np.zeros((npts, npts))
            pot2[row_inds, col_inds] = pot_contribs
            pot2[col_inds, row_inds] = pot_contribs
            pot += pot2

        return pot

    @classmethod
    def evaluate_multiplicative(
            cls,
            function,
            overlap_data:'OverlapGaussianData',
            integral_handler=None,
            expansion_degree=None,
            expansion_type=None,
            quadrature_degree=None,
            pairwise_functions=None,
            logger=None
    ):

        if integral_handler is None:
            if isinstance(function, dict):
                if 'analytic_integrals' in function:
                    integral_handler = 'analytic'
            elif expansion_degree is not None:
                integral_handler = 'expansion'
            else:
                integral_handler = 'quad'

        if not isinstance(logger, Logger):
            logger = Logger.lookup(logger)


        if integral_handler == 'quad':
            if pairwise_functions is not None:
                raise ValueError("pairwise functions can't go with direct quadrature")
            logger.log_print("evaluating integrals with {n}-order quadrature about {npt} points",
                             n=quadrature_degree,
                             npt=overlap_data.npts
                             )
            pot_mat = cls.quad_integrate(
                function,
                overlap_data,
                degree=quadrature_degree,
                logger=logger
            )
        elif integral_handler == 'expansion':
            logger.log_print("evaluating integrals with {n}-degree expansions about {npt} points",
                             n=expansion_degree,
                             npt=overlap_data.npts
                             )
            pot_mat = cls.expansion_integrate(function,
                                              overlap_data,
                                              expansion_degree=expansion_degree,
                                              expansion_type=expansion_type,
                                              pairwise_functions=pairwise_functions,
                                              logger=logger
                                              )
        elif integral_handler == 'analytic':
            logger.log_print("evaluating integrals analytically", n=expansion_degree)
            pot_mat = cls.analytic_integrate()
        else:
            raise ValueError("unknown operator evaluation scheme {}".format(integral_handler))

        return pot_mat

    def evaluate_pe(self, overlap_data:'OverlapGaussianData', logger=None):
        if logger is None:
            logger = self.logger
        logger.log_print("evaluating potential energy contribution")
        return self.evaluate_multiplicative(
            self.potential_function,
            overlap_data,
            integral_handler=self.handler,
            quadrature_degree=self.quadrature_degree,
            expansion_degree=self.expansion_degree,
            expansion_type=self.expansion_type,
            pairwise_functions=self.pairwise_handler,
            logger=logger
        )

    def evaluate_op(self,
                    operator,
                    overlap_data:'OverlapGaussianData',
                    integral_handler=None,
                    expansion_degree=None,
                    expansion_type=None,
                    quadrature_degree=None,
                    pairwise_functions=None,
                    logger=None
                    ):
        no_type_passed = expansion_degree is None and quadrature_degree is None
        return self.evaluate_multiplicative(
            operator,
            overlap_data,
            integral_handler=self.handler if integral_handler is None else integral_handler,
            quadrature_degree=self.quadrature_degree if no_type_passed else quadrature_degree,
            expansion_degree=self.expansion_degree if no_type_passed else expansion_degree,
            expansion_type=self.expansion_type if expansion_type is None else expansion_type,
            pairwise_functions=pairwise_functions,
            logger=self.logger if logger is None else logger
        )

class DGBPairwisePotentialEvaluator(DGBEvaluator, metaclass=abc.ABCMeta):
    def __init__(self, coords, pairwise_potential_functions, quadrature_degree=3,
                 use_with_interpolation='ignored'
                 ):
        self.coords = coords
        self.pairwise_potential_functions = pairwise_potential_functions
        self.quadrature_degree = quadrature_degree

    @classmethod
    def get_bond_length_deltas(cls, natoms, ndim, i, j, full=False):
        if not full:
            mat = np.zeros((ndim, natoms, ndim))
            for k in range(ndim):
                mat[k][i][k] = 1
                mat[k][j][k] = -1
            mat = mat.reshape((mat.shape[0], natoms * ndim))
        else:
            mat = np.sqrt(2) * np.eye(natoms*ndim).reshape((natoms, ndim, natoms, ndim))
            for k in range(ndim):
                mat[i][k][i][k] = 1
                mat[i][k][j][k] = -1
                mat[j][k][i][k] = 1
                mat[j][k][j][k] = 1
            mat = mat.reshape((natoms * ndim, natoms * ndim))
        return mat

    @abc.abstractmethod
    def get_coordinate_bond_length_projection(self, i, j) -> 'tuple[np.ndarray, np.ndarray]':
        ...
    def get_coordinate_change_transformation(self, coordinate_projection_data) -> np.ndarray:
        v, s, w = np.linalg.svd(coordinate_projection_data)  # w is the important transformation
        # we get the dimension for every specified transformation (doesn't always need to be the same)
        if not np.allclose(np.abs(s), s):
            s_sorts = np.argsort(-np.abs(s), axis=-1)
            s = np.abs(s[s_sorts])
            w = w[s_sorts]
        w = w.transpose(0, 2, 1)
        dims = np.sum(np.abs(s) > 1e-14, axis=1)
        udims, upos = np.unique(dims, return_index=True)
        if len(udims) == 1:
            return w[:, :, :udims[0]]
        else:
            raise NotImplementedError("bond length transforms with different dimensions not yet supported...")

    def get_bond_length_change_transformation(self, overlap_data:'OverlapGaussianData', i, j) -> np.ndarray:
        d0, base_proj = self.get_coordinate_bond_length_projection(i, j)
        proj_data = np.broadcast_to(
            base_proj[np.newaxis],
            (len(overlap_data.rotations),) + base_proj.shape
        ) #@ overlap_data.rotations.transpose(0, 2, 1) # not sure if we need a .T or not...
        return self.get_coordinate_change_transformation(proj_data)

    def wrap_distance_function(self, i, j, overlap_data:'OverlapGaussianData', transformations, pairwise_function):
        d0, proj = self.get_coordinate_bond_length_projection(i, j) # dim x ncoordx
        tf_proj = proj[np.newaxis] @ transformations # N x dim x subdim
        @functools.wraps(pairwise_function)
        def eval_pairwise(coords, *, deriv_order=None, d0=d0, tf_proj=tf_proj):
            # coords is expressed in terms of some sort of transformed version
            # of the normal modes (or Cartesians) that were supplied to the program
            # and we know exactly how these change the deltas that feed into the bond
            # length using tf_proj from above, this means we can get the displacements of the deltas
            # for free and ... add these onto the initial deltas???
            base_shape = coords.shape[:-1]
            coords = coords.reshape(tf_proj.shape[0], -1, coords.shape[-1]) # N x k x subdim
            proj_d = tf_proj[:, np.newaxis, :, :] @ coords[:, :, :, np.newaxis] # N x k x dim x 1
            d = d0[np.newaxis, np.newaxis, :] + proj_d.reshape(proj_d.shape[:3])

            if d.shape[1] == 1:
                d = d.reshape((d.shape[0], d.shape[2]))

            if deriv_order is None or deriv_order == 0:
                r = np.linalg.norm(d, axis=-1)  # we ensure no scaling of difference coords
                r_derivs = None
            else:
                r_derivs = nput.vec_norm_derivs(d, order=deriv_order)
                r, r_derivs = r_derivs[0], r_derivs[1:]

            fdat = pairwise_function(r, deriv_order=deriv_order)
            if r_derivs is not None and len(r_derivs) > 0:
                # first we get df/d_delta by transforming df/dr.dr/d_delta
                f_vals = fdat[0]
                r_derivs = [r[..., np.newaxis] for r in r_derivs]
                f_derivs = [fdat[i].reshape(f_vals.shape + (1,) * i) for i in range(1, deriv_order + 1)]
                f_derivs = list(TensorDerivativeConverter(r_derivs, f_derivs).convert())
                # why bother? -> we're just gonna remove this part of the transformation anyway
                # then we get df/dq by transforming with the linear transformation d_del/dq
                q_derivs = []
                for i,d in enumerate(f_derivs):
                    for j in range(d.ndim-1):
                        d = np.tensordot(d, proj, axes=[1, 0])
                        # d = nput.vec_tensordot(d, tf_proj, shared=1, axes=[1, 1])
                    q_derivs.append(d)
                fdat = [f_vals] + q_derivs

            if deriv_order is not None:
                # print(base_shape, [f.shape for f in fdat])
                fdat = [f.reshape(base_shape + f.shape[r.ndim:]) for f in fdat]
            else:
                fdat = fdat.reshape(base_shape + fdat.shape[r.ndim:])
            return fdat

        return eval_pairwise

    # def get_angle_change_transformation(self, coords, i, j, k):
    #     ...

    # def wrap_angle_function(self, coord, i, j, k, transformations, pairwise_function):
    #     ...

    def evaluate_pairwise_contrib(self,
                                  overlap_data:'OverlapGaussianData',
                                  quadrature_degree=None,
                                  expansion_degree=2
                                  ):

        if quadrature_degree is None:
            quadrature_degree = self.quadrature_degree

        centers = overlap_data.centers
        covs = overlap_data.covariance_matrices
        mom_sum = overlap_data.momenta_sum
        mom_diff = overlap_data.momenta_diff

        npts = centers.shape[0]
        fdim = centers.shape[1]

        potential_contrib = 0
        if expansion_degree is not None:
            derivative_contribs = [
                np.zeros((npts,) + (fdim,) * i)
                    if i > 0 else
                np.zeros((npts,))
                for i in range(expansion_degree + 1)
            ]
        else:
            derivative_contribs = [np.zeros((npts,))]

        for index_pair, pairwise_func in self.pairwise_potential_functions.items():
            if len(index_pair) == 2:
                i,j = index_pair
            else:
                raise NotImplementedError("currently only pairwise distances potentials are supported")

            tfsT = self.get_bond_length_change_transformation(overlap_data, i, j)
            tfs = tfsT.transpose((0, 2, 1)) # tfs is unitary so this is the inverse
            f = self.wrap_distance_function(i, j, overlap_data, tfsT, pairwise_func)
            sub_cov = tfs @ covs @ tfsT
            tf_cov = np.linalg.inv(sub_cov)

            tf_alphas, tf_vecs = np.linalg.eigh(tf_cov)
            tf_alphas = tf_alphas / 2 # baked into the covariances...
            ndim = tf_alphas.shape[1]

            tf_centers = tfs @ centers[:, :, np.newaxis]
            tf_centers = tf_centers.reshape(tf_centers.shape[:2])

            if mom_sum is not None:
                tf_momenta = [
                    np.reshape(tf_cov @ tfs @ covs @ mom_diff[:, :, np.newaxis], sub_cov.shape[:2]),
                    np.reshape(tf_cov @ tfs @ covs @ mom_sum[:, :, np.newaxis], sub_cov.shape[:2]),
                ]
            else:
                tf_momenta = None

            pairwise_contrib = self.rotated_gaussian_quadrature(
                f,
                tf_alphas,
                tf_centers,
                tf_vecs,
                tf_vecs.transpose(0, 2, 1),
                tf_momenta,
                degree=quadrature_degree,
                normalize=False
            )

            potential_contrib += pairwise_contrib

            if expansion_degree is not None:
                # we make use of the inverse of the transformation that takes us from
                # the space of total coordinates (normal modes or Cartesians) to the space
                # of coordinates (deltas) that lead to a change in the coordinate of interest
                # we'll note that our wrapped function is expected to implement the appropriate
                # conversion from derivatives with respect to r to derivatives with respect to
                # the deltas

                tf_derivs_contribs = f(tf_centers, deriv_order=expansion_degree)
                for i, d in enumerate(tf_derivs_contribs):
                    if i == 0:
                        derivative_contribs[0] += d
                    else:
                        # if i == 2:
                        #     print("pre", d[0])
                        # for j in range(i):
                        #     d = nput.vec_tensordot(d, tfsT, shared=1, axes=[1, 2])
                        # if i == 2:
                        #     print("post", d[0])
                        derivative_contribs[i] += d
            else:
                derivative_contribs[0] += f(tf_centers[:, :ndim])

        return potential_contrib, derivative_contribs

class DGBCartesianPairwiseEvaluator(DGBPairwisePotentialEvaluator):

    def __init__(self, coords:'DGBCartesians', pairwise_functions, **opts):
        super().__init__(coords, pairwise_functions, **opts)

    def get_coordinate_bond_length_projection(self, i, j):
        natoms, ndim = self.coords.cart_shape[1:]
        base_mat = self.get_bond_length_deltas(natoms, ndim, i, j)
        # inverse_projection = base_mat.T / 2
        d0 = np.zeros(ndim)
        return d0, base_mat

class DGBWatsonPairwiseEvaluator(DGBPairwisePotentialEvaluator):
    def __init__(self, coords:'DGBWatsonModes', pairwise_functions, **opts):
        super().__init__(coords, pairwise_functions, **opts)

    def get_coordinate_bond_length_projection(self, i, j):
        natoms = len(self.coords.masses)
        modes = self.coords.modes.matrix
        ndim = modes.shape[0] // natoms # this will almost always be 3
        base_mat = self.get_bond_length_deltas(natoms, ndim, i, j)
        tf_base = base_mat @ modes
        d0 = np.dot(base_mat, self.coords.modes.origin.reshape(-1))
        return d0, tf_base




"""
Molecules provides wrapper utilities for working with and visualizing molecular systems
"""

from .Vibrations import *
from .Molecule import *
from .CoordinateSystems import *

# getting the full list of symbols explicitly in an __all__ variable
__all__ = []
from .Vibrations import __all__ as exposed
__all__ += exposed
from .Molecule import __all__ as exposed
__all__ += exposed
from .Properties import *; from .Properties import __all__ as exposed
__all__ += exposed
from .CoordinateSystems import __all__ as exposed
__all__ += exposed


"""
Provides a DVRWavefunction class that inherits from the base Psience wavefunction
"""

import numpy as np

from McUtils.Plots import Graphics
from McUtils.Zachary import Interpolator

from Psience.Wavefun import Wavefunction, Wavefunctions
from .BaseDVR import DVRResults

__all__ = ["DVRWavefunctions", "DVRWavefunction"]

class DVRWavefunction(Wavefunction):

    def __init__(self, energy, data, parent=None, grid=None, index=None, **opts):
        super().__init__(energy, data, parent=parent, index=index, **opts)
        if grid is None:
            grid = self.parent.grid
        self.grid = grid
        self._interp = None
    def get_dimension(self):
        return self.grid.shape[-1]

    def plot(self, figure=None, grid=None, **opts):
        if grid is None:
            grid = self.grid
        return super().plot(figure=figure, grid=grid, values=self.data, **opts)

    def expectation(self, op, other=None):
        """Computes the expectation value of operator op over the wavefunction other and self

        :param other:
        :type other: Wavefunction | np.ndarray
        :param op:
        :type op:
        :return:
        :rtype:
        """
        wf = op(self.data)
        if not isinstance(other, np.ndarray):
            other = other.data
        return np.dot(wf, other)

    @property
    def interp(self):
        if self._interp is None:
            self._interp = Interpolator(self.grid, self.data)
        return self._interp
    def evaluate(self, points):
        """
        Evaluates the functions at the given points

        :return:
        :rtype:
        """
        return self.interp(points)

    def marginalize_out(self, dofs):
        """
        Computes the projection of the current wavefunction onto a set of degrees
        of freedom

        :return:
        :rtype:
        """

        # if isinstance(dofs, (int, np.integer)):
        #     dofs = [dofs]
        # dofs = np.flip(np.sort(dofs))

        raise NotImplementedError("DVR projections not yet implemented")

class DVRWavefunctions(Wavefunctions):
    # most evaluations are most efficient done in batch for DVR wavefunctions so we focus on the batch object
    wavefunction_class = DVRWavefunction
    def __init__(self, energies=None, wavefunctions=None, grid=None, results:DVRResults=None, **opts):
        super().__init__(energies=energies, wavefunctions=wavefunctions, results=results, grid=grid, **opts) # add all opts
        self.results = results
        self.grid = grid
    def __repr__(self):
        return "{}(num={}, DVR={})".format(
            type(self).__name__,
            len(self),
            self.results.parent
        )

    # def __len__(self):
    #     return len(self.energies)
    # def __iter__(self):
    #     for i in range(len(self)):
    #         yield self.__getitem__(i)
    # def __getitem__(self, item):
    #     """
    #     Provides a single `DVRWavefunction` or slice of `DVRWavefunctions`
    #     :param item:
    #     :type item:
    #     :return:
    #     :rtype: DVRWavefunction | DVRWavefunctions
    #     """
    #     if not isinstance(item, (int, np.integer)):
    #         return type(self)(
    #             energies=self.energies[item],
    #             wavefunctions=self.wavefunctions[:, item],#.reshape((len(self.wavefunctions), -1)), # WTF? Was I transposing here?...?
    #             wavefunction_class=self.wavefunction_class,
    #             results=self.results,
    #             grid=self.grid,
    #             **self.opts
    #         )
    #     else:
    #         return self.wavefunction_class(
    #             self.energies[item],
    #             self.wavefunctions[:, item],
    #             parent=self,
    #             **self.opts
    #         )

    def plot(self, figure=None, **opts):
        """
        Plots the held wavefunctions

        :param figure:
        :type figure:
        :param graphics_class:
        :type graphics_class:
        :param plot_style:
        :type plot_style:
        :param scaling:
        :type scaling:
        :param shift:
        :type shift:
        :param opts:
        :type opts:
        :return:
        :rtype: Graphics
        """

        grid = self.grid

        dim = len(grid.shape) # mesh grid...
        if dim > 1 and grid.shape[-1] == dim-1: # check whether we have a mesh of points that we need to reshape
            grid = np.moveaxis(grid, grid.ndim, 0)

        return super().plot(
            figure=figure,
            grid=grid,
            **opts
        )

    def expectation(self, op, other=None, multiplicative=True):
        """Computes the expectation value of operator op over the wavefunction other and self

        :param other:
        :type other: DVRWavefunctions | np.ndarray
        :param op:
        :type op:
        :return:
        :rtype:
        """
        if other is None:
            other = self
        if not multiplicative:
            raise ValueError("don't have non-multiplicative operators supported yet...")
        else:
            if not isinstance(op, np.ndarray):
                op = op(self.grid.reshape((-1,) + self.grid.shape[2:]))
            wfs = self.wavefunctions
            for _ in range(op.ndim-1):
                wfs = np.expand_dims(wfs, -1)
            wfs = np.expand_dims(op, 1) * wfs
            # print(self.wavefunctions.shape, wfs.shape)
        if not isinstance(other, np.ndarray):
            other = other.wavefunctions
        ev = np.tensordot(other, wfs, axes=[0, 0])
        ev = ev.transpose([1, 0] + list(range(2, ev.ndim)))
        # print("--->", wfs.shape, other.shape, ev.shape)
        return ev

    def transform_operator(self, M):
        if hasattr(M, 'toarray'):
            M = M.toarray()
        return np.dot(np.dot(self.wavefunctions.T, M), self.wavefunctions)

    def coordinate(self):
        return self.expectation(self.results.grid)
    def momentum(self):
        dvr = self.results.parent
        p = dvr.real_momentum(grid=self.results.grid, **dvr.opts)
        return self.transform_operator(p)
    def laplacian(self):
        dvr = self.results.parent
        res = dvr.run(mass=1, g=None, hb=1, potential_function=lambda g:np.zeros(len(g)), result='kinetic_energy')
        p2 = -2*res.kinetic_energy
        return self.transform_operator(p2)
    def kinetic_energy(self):
        # import McUtils.Plots as plt
        # plt.ArrayPlot(self.results.kinetic_energy)
        # plt.ArrayPlot(self.transform_operator(self.results.kinetic_energy)).show()
        # print(self.results.kinetic_energy)
        return self.transform_operator(self.results.kinetic_energy)
    def potential_energy(self):
        return self.transform_operator(self.results.kinetic_energy)


"""
Provides very general support for an abstract wavefunction object
Allows different methods to provide their own concrete implementation details
"""
import abc
from abc import *
import numpy as np

from McUtils.Plots import Graphics, Plot, TriContourPlot
from ..Spectra import DiscreteSpectrum

__all__ = [
    "Wavefunction",
    "Wavefunctions",
    "WavefunctionException"
]

__reload_hook__ = ["..Spectra"]

class WavefunctionException(Exception):
    pass

class Wavefunction:
    """Represents a single wavefunction object"""
    def __init__(self, energy, data, parent=None, index=None, **opts):
        self.energy = energy
        self.data   = data
        self.parent = parent
        self.index = index
        self.opts   = opts

    @abc.abstractmethod
    def get_dimension(self):
        raise NotImplementedError("abstract base method")
    @property
    def ndim(self):
        return self.get_dimension()

    @classmethod
    def prep_plot_grid(cls,
                       domain,
                       plot_points=100,
                       domain_padding=None
                       ):
        if isinstance(domain[0], (int, np.integer, float, np.floating)):
            domain = [domain]
        if isinstance(plot_points, (int, np.integer)):
            plot_points = [plot_points] * len(domain)

        if domain_padding is not None:
            if isinstance(domain_padding, (int, float, np.integer, np.floating)):
                domain_padding = [-domain_padding, domain_padding]
            domain_padding = np.asanyarray(domain_padding)
            if domain_padding.ndim == 1:
                domain_padding = domain_padding[np.newaxis, :]
            domain = np.asanyarray(domain) + domain_padding

        grids = []
        for dom, pts in zip(domain, plot_points):
            grids.append(np.linspace(*dom, pts))
        grid = np.moveaxis(np.array(np.meshgrid(*grids, indexing='xy')), 0, -1).reshape(-1, len(domain))  # vector of points

        return grid

    def plot(self,
             figure=None, domain=None, *, domain_padding=None, grid=None, values=None, plot_points=100,
             which=None,
             index=0, scaling=1, shift='auto', plotter=None, plot_density=False,
             zero_tol=1e-8, contour_levels=None,
             **opts
             ):
        """
        Plots a single wave function on the grid

        :param figure:
        :type figure:
        :param grid:
        :type grid:
        :param index:
        :type index:
        :param scaling:
        :type scaling:
        :param shift:
        :type shift:
        :param opts:
        :type opts:
        :return:
        :rtype:
        """
        if which is None:
            which = index

        if grid is None and domain is None:
            raise ValueError("can't plot a wave function without a specified domain")

        if grid is None:
            grid = self.prep_plot_grid(domain=domain, domain_padding=domain_padding, plot_points=plot_points)

        grid = np.asanyarray(grid)
        if grid.ndim == 1:
            grid = grid[:, np.newaxis]
        elif grid.ndim > 2:
            grid = grid.reshape(-1, grid.shape[-1])
        dim = grid.shape[-1]

        if dim > 2 and plotter is None: # if people want to try, let 'em
            raise ValueError("can't plot data with dimension higher than 2, take a projection first")

        # allows us to scale wave functions independently
        if not isinstance(scaling, (int, float, np.integer, np.floating)):
            scaling = scaling[which]
        if shift is None: shift = 0
        if isinstance(shift, str) and shift == 'auto':
            shift = self.energy
        if not isinstance(shift, (int, float, np.integer, np.floating, str)):
            shift = shift[which]

        if values is None:
            if plot_density:
                values = self.probability_density(grid)
            else:
                values = self.evaluate(grid)
        values[np.abs(values) < zero_tol] = 0.

        values = values * scaling
        if contour_levels is not None and 'levels' not in opts:
            if np.std(values) > 1e-8:
                max_val = np.max(np.abs(values))
                levels = np.linspace(-max_val + shift, max_val + shift, contour_levels)
                opts['levels'] = levels
        values = values + shift

        if plotter is None:
            if dim == 1:
                plotter = Plot
            else:
                plotter = TriContourPlot

        return plotter(*np.moveaxis(grid, -1, 0), values, figure=figure, **opts)

    def projection_plot(self,
                        coords,
                        figure=None,
                        **plot_options
                        ):
        """
        A convenience function to plot multiple projections
        on the same set of axes

        Deprecated in favor of `plot_cartesians` for its primary use case (`DGBWavefunctions`)

        :param coords:
        :type coords:
        :param figure:
        :type figure:
        :param plot_options:
        :type plot_options:
        :return:
        :rtype:
        """
        if isinstance(coords, (int, np.integer)):
            coords = [[coords]]
        elif isinstance(coords[0], (int, np.integer)):
            coords = [[c] for c in coords]

        for proj_inds in coords:
            fig = self.project(proj_inds).plot(
                figure=figure,
                **plot_options
            )
            if figure is None:
                figure = fig

        return figure

    @abstractmethod
    def expectation(self, op, other=None):
        """Computes the expectation value of operator op over the wavefunction other and self

        :param other:
        :type other: Wavefunction
        :param op:
        :type op:
        :return:
        :rtype:
        """
        pass
    def overlap(self, other):
        return self.expectation(lambda w:w, other=other)
    @abstractmethod
    def evaluate(self, points):
        """
        Evaluates the current wavefunction

        :return:
        :rtype:
        """
        raise NotImplementedError("abstract base method")
    @property
    def probability_density(self):
        """
        Computes the probability density of the current wavefunction

        :return:
        :rtype:
        """
        return lambda pts:self.evaluate(pts)**2 # we're assuming real I guess...
    @abstractmethod
    def marginalize_out(self, dofs):
        """
        Integrates out the contributions from the degrees of freedom `dofs`

        :return:
        :rtype: Wavefunction
        """
        raise NotImplementedError("abstract base method")
    def project(self, dofs):
        """
        Computes the projection of the current wavefunction onto a set of degrees
        of freedom, returning a projected wave function object

        :return:
        :rtype: Wavefunction
        """

        dof_complement = np.setdiff1d(np.arange(self.ndim), dofs)
        return self.marginalize_out(dof_complement)

class Wavefunctions:
    """
    An object representing a set of wavefunctions.
    Provides concrete, but potentially inefficient methods for doing all the wavefunction ops.

    """
    wavefunction_class = Wavefunction
    def __init__(self,
                 energies=None, wavefunctions=None,
                 indices=None, wavefunction_class=None,
                 dipole_function=None,
                 **opts):
        self.wavefunctions = wavefunctions
        self.energies = energies
        self.wavefunction_class = self.wavefunction_class if wavefunction_class is None else wavefunction_class
        self.indices = indices
        self.dipole_function = dipole_function
        self.opts = opts

    def get_wavefunctions(self, which):
        inds = self.indices
        if inds is None:
            inds = np.arange(len(self.wavefunctions))
        if not isinstance(which, (int, np.integer)):
            return type(self)(
                energies=self.energies[which,],
                wavefunctions=self.wavefunctions[:, which],
                wavefunction_class=self.wavefunction_class,
                indices=inds[which,],
                dipole_function=self.dipole_function,
                **self.opts
            )
        else:
            return self.wavefunction_class(
                self.energies[which],
                self.wavefunctions[:, which],
                parent=self,
                index=inds[which],
                **self.opts
            )
    def __getitem__(self, item):
        """Returns a single Wavefunction object"""
        # iter comes for free with this
        return self.get_wavefunctions(item)
    def __len__(self):
        return len(self.energies)
    def __iter__(self):
        for i in range(len(self)):
            yield self.__getitem__(i)

    def frequencies(self, start_at = 0):
        return np.concatenate([self.energies[:start_at], self.energies[1+start_at:]]) - self.energies[start_at]

    def get_spectrum(self,
                     dipole_function=None,
                     *,
                     start_at=0,
                     **options
                     ):
        if dipole_function is None: # it's just so convenient to have this on the object...
            dipole_function = self.dipole_function
        if dipole_function is None:
            raise ValueError("a dipole function is required to get a spectrum (none stored in wavefunctions)")
        freqs = self.frequencies(start_at=start_at)
        transition_moments = self.expectation(dipole_function,
                                              **options,
                                              )[start_at]
        transition_moments = np.concatenate([
            transition_moments[:start_at],
            transition_moments[start_at+1:]
        ])
        return DiscreteSpectrum.from_transition_moments(
            freqs,
            transition_moments
        )

    def plot(self, figure=None, graphics_class=None, **opts):
        """Plots all of the wavefunctions on one set of axes

        :param opts:
        :type opts:
        :return:
        :rtype:
        """

        k = "plot_defaults"
        opts = dict(self.opts[k] if k in self.opts else (), **opts)

        if figure == None:
            dim = self.opts['dimension'] if 'dimension' in self.opts else 1
            if graphics_class is None:
                if dim ==1:
                    graphics_class = Graphics
                elif dim == 2:
                    graphics_class = Graphics#Graphics3D
                else:
                    raise WavefunctionException(
                        "{}.{}: don't know how to plot wavefunctions of dimension {}".format(
                            type(self).__name__, 'plot', dim
                        )
                    )
            figure = graphics_class(strict=False, **opts)

        for i, wfn in enumerate(self):
            ind = wfn.index
            if ind is None:
                ind = i
            wfn.plot(figure, which=i, index=ind, **opts)

        return figure

    def expectation(self, op, other=None):
        """
        Computes the expectation value of operator op over the wavefunction other and self

        :param other:
        :type other: Wavefunctions
        :param op:
        :type op:
        :return:
        :rtype:
        """
        if other is None:
            other = self

        res = []
        for wfn in self:
            subres = []
            for ofn in other:
                subres.append(wfn.expectation(op, ofn))
            res.append(subres)
        return np.array(res)
    def overlap(self, other):
        return self.expectation(lambda w:w, other=other)

    def coordinate(self):
        """
        Provides the coordinate operator in the wavefunction basis

        :return:
        :rtype:
        """
        raise NotImplementedError("no coordinate rep implemented for {}".format(self))
    def momentum(self):
        """
        Provides the real part of the representation of the momentum operator in the wavefunction basis

        :return:
        :rtype:
        """
        raise NotImplementedError("no momentum implemented for {}".format(self))
    def laplacian(self):
        """
        Provides the representation of the laplacian in the wavefunction basis

        :return:
        :rtype:
        """
        raise NotImplementedError("no momentum implemented for {}".format(self))
    def kinetic_energy(self):
        """
        Provides the representation of the KE in the wavefunction basis

        :return:
        :rtype:
        """
        raise NotImplementedError("no KE implemented for {}".format(self))


"""
Simple utilities for doing basic AIMD simulations as well
as utilities for managing AIMD sims
"""

__all__ = []
from .Simulator import *; from .Simulator import __all__ as exposed
__all__ += exposed


import itertools

import numpy as np, scipy, math

from McUtils.Data import AtomData, UnitsData
import McUtils.Numputils as nput
from .Helpers import AnalyticModelBase
from .AnalyticJacobianDotCalculator import InternalJacobianDisplacements
from ..Data import KEData

__all__ = [
    'AnalyticPotentialConstructor',
    'AnalyticKineticEnergyConstructor',
    'AnalyticModel',
    "MolecularModel",
    "GeometricFunction"
]
__reload_hook__ = ["..Data"]

class AnalyticPotentialConstructor(AnalyticModelBase):
    """
    Provides a set of symbolic potentials for use in models

    :related:AnalyticModel, AnalyticKineticEnergyConstructor
    """
    @classmethod
    def morse(cls, i, j, De=None, a=None, re=None, eq=None, w=None, wx=None):
        """
        Returns a fully symbolic form of a Morse potential
        :return:
        :rtype:
        """

        if De is None and w is not None:
            if isinstance(w, str):
                w = AnalyticModelBase.symbol(w, i, j)
            if isinstance(wx, str):
                wx = AnalyticModelBase.symbol(wx, i, j)
            muv = (
                    1/AnalyticModelBase.symbolic_m(i)
                    + 1/AnalyticModelBase.symbolic_m(j)
            )
            De = (w ** 2) / (4 * wx)
            a = AnalyticModelBase.sym.sqrt(2 * wx / muv)

        if re is None:
            re = eq

        return cls.calc_morse(
            AnalyticModelBase.symbol("De", i, j) if De is None else De,
            AnalyticModelBase.symbol("ap", i, j) if a is None else a,
            AnalyticModelBase.symbolic_r(i, j),
            AnalyticModelBase.symbol("re", i, j) if re is None else re
        )
    @staticmethod
    def calc_morse(De, a, r, re):
        return De * (1 - AnalyticModelBase.sym.exp(-a * (r-re)))**2

    @staticmethod
    def harm(k, x, x_e):
        return k*(x-x_e)**2
    @classmethod
    def harmonic(cls, *args, k=None, eq=None, qe=None):
        """
        Returns a fully symbolic form of a Morse potential
        :return:
        :rtype:
        """

        if qe is None:
            qe = eq

        return cls.harm(
            AnalyticModelBase.symbol("k", *args) if k is None else k,
            AnalyticModelBase.var(*args),
            AnalyticModelBase.symbol("qe", *args) if qe is None else qe,
        )


    @staticmethod
    def lin(k, x, x_e):
        return k * (x - x_e)
    @classmethod
    def linear(cls, *args, k=1, eq=None, xe=None):
        """
        Returns a fully symbolic form of a linear function
        :return:
        :rtype:
        """
        if xe is None:
            xe = eq
        return cls.lin(
            AnalyticModelBase.symbol("k", *args) if k is None else k,
            AnalyticModelBase.var(*args),
            AnalyticModelBase.symbol("xe", *args) if xe is None else xe
        )

    @staticmethod
    def pow(k, x, x_e, n):
        return k * (x - x_e) ** n
    @classmethod
    def power(cls, *args, k=1, eq=None, n=None, xe=None):
        """
        Returns a fully symbolic form of a linear function
        :return:
        :rtype:
        """
        if xe is None:
            xe = eq
        return cls.pow(
            AnalyticModelBase.symbol("k", *args) if k is None else k,
            AnalyticModelBase.var(*args),
            AnalyticModelBase.symbol("xe", *args) if xe is None else xe,
            AnalyticModelBase.symbol("n", *args) if n is None else n
        )

    @classmethod
    def cos(cls, *args, eq=None, qe=None):
        """
        Returns a fully symbolic form of cos
        :return:
        :rtype:
        """
        if qe is None:
            qe = eq
        return AnalyticModelBase.sym.cos(
            AnalyticModelBase.var(*args) -
                (AnalyticModelBase.symbol("qe", *args) if qe is None else qe)
        )

    @classmethod
    def sin(cls, *args, eq=None, qe=None):
        """
        Returns a fully symbolic form of sin
        :return:
        :rtype:
        """
        if qe is None:
            qe = eq
        return AnalyticModelBase.sym.sin(
            AnalyticModelBase.var(*args) -
            (AnalyticModelBase.symbol("qe", *args) if qe is None else qe)
        )

    @classmethod
    def multiwell(cls, *args, turning_points=None, origin=None, eq=None, minimum=0, depth=None):
        """

        :param args:
        :type args:
        :param turning_points:
        :type turning_points:
        :param depth:
        :type depth:
        :return:
        :rtype:
        """

        if turning_points is None:
            raise ValueError("need turning points for multiwell")
        if origin is None:
            origin = eq
        if origin is None:
            origin = 0
        turning_points = [origin+t for t in turning_points]
        x = AnalyticModelBase.var(*args)
        D = AnalyticModelBase.symbol("D", *args) if depth is None else depth
        terms = [x-t for t in turning_points]
        integrand = terms[0]
        for t in terms[1:]:
            integrand *= t
        poly = AnalyticModelBase.sym.integrate(integrand, x)
        min_val = np.inf
        min_pos = None
        turning_vals = [poly.subs([[x, t]]) for t in turning_points]
        for n, v in enumerate(turning_vals):
            if v < min_val:
                min_val = v
                min_pos = n
        if min_pos == 0:
            barrier_height = turning_vals[min_pos+1] - turning_vals[min_pos]
        elif min_pos == len(turning_vals) - 1:
            barrier_height = turning_vals[min_pos-1] - turning_vals[min_pos]
        else:
            barrier_height = min(turning_vals[min_pos+1], turning_vals[min_pos-1]) - turning_vals[min_pos]
        return D/barrier_height * (poly - min_val) + minimum

class GeometricFunction:
    def __init__(self, coords, coord_funcs, val_func):
        self.coords = coords
        self.coord_funcs = coord_funcs
        self.val_func = val_func
    def __call__(self, masses, coords):
        coord_vals = [cf(masses, coords) for cf in self.coord_funcs]
        return self.val_func(coord_vals)

    @classmethod
    def position_function(cls, i):
        return lambda masses, coords: coords[..., i, :]
    @classmethod
    def normal_function(cls, i, j, k):
        return lambda masses, coords: nput.vec_crosses(
            coords[..., j, :] - coords[..., i, :],
            coords[..., k, :] - coords[..., j, :],
            normalize=True
        )
    @classmethod
    def mass_function(cls, i):
        return lambda masses, coords: np.full(coords.shape[:-2], masses[i])
    @classmethod
    def distance_function(cls, i, j):
        return lambda masses, coords: nput.pts_norms(coords[..., i, :], coords[..., j, :])
    @classmethod
    def angle_function(cls, i, j, k):
        return lambda masses, coords: nput.pts_angles(coords[..., i, :], coords[..., j, :], coords[..., k, :],
                                                      return_crosses=False,
                                                      return_norms=False
                                                      )
    @classmethod
    def dihedral_function(cls, i, j, k, l):
        return lambda masses, coords: nput.pts_dihedrals(coords[..., i, :], coords[..., j, :], coords[..., k, :], coords[..., l, :])
    @classmethod
    def book_function(cls, i, j, k, l):
        return lambda masses, coords: nput.pts_book_angles(coords[..., i, :], coords[..., j, :], coords[..., k, :], coords[..., l, :])

    @classmethod
    def create_symbol_function(cls, sym):
        sym_type, inds = AnalyticModel.parse_symbol(sym)
        func_type = {
            'x':cls.position_function,
            'n':cls.normal_function,
            'm':cls.mass_function,
            'r':cls.distance_function,
            'a':cls.angle_function,
            't':cls.dihedral_function,
            'y':cls.book_function
        }
        return func_type.get(sym_type)(*inds)
    @classmethod
    def sorted_vars(cls, vars):
        parse = {
            sym:AnalyticModel.parse_symbol(sym)
            for sym in vars
        }
        return list(sorted(
            vars,
            key=lambda sym:[
                len(parse[sym][1]),
                sorted(parse[sym][1]),
                parse[sym][0]
                ]
        ))

    @classmethod
    def from_expr(cls, expr):
        if nput.is_numeric(expr):
            coord_funcs = [lambda masses, coords: np.empty(coords.shape[:-2])]
            val_func = lambda coords, _val=expr: np.full(coords[0].shape, _val)
            return cls(None, coord_funcs, val_func)
        else:
            base_coords = cls.sorted_vars(expr.free_symbols)
            sym_evals = [cls.create_symbol_function(sym) for sym in base_coords]
            val_func = AnalyticModel.lambdify(base_coords, expr, dummify=True, rewrite_trig=True, lambdify_arrays=True)
            return cls(base_coords, sym_evals, val_func)

class AnalyticKineticEnergyConstructor(AnalyticModelBase):
    """
    Provides G and V' elements from Frederick and Woywood

    :related:AnalyticModel, AnalyticPotentialConstructor
    """

    @classmethod
    def _remap_inter_inds(cls, inds, intersection, remapping, n_inter, n_diff):
        _ = []
        for i in inds:
            if i not in remapping:
                if i in intersection:
                    remapping[i] = n_inter
                    _.append(n_inter)
                    n_inter += 1
                else:
                    remapping[i] = n_diff
                    _.append(n_diff)
                    n_diff += 1
            else:
                _.append(remapping[i])
        return tuple(_), n_inter, n_diff

    @classmethod
    def _canonicalize_inds(cls, t, inds):
        if t == 'r':
            i,j = inds
            if j < i:
                return (j,i), (1, 0)
        elif t == 'a':
            i,j,k = inds
            if k < i:
                return (k,j,i), (2, 1, 0)
        elif t == 't':
            i,j,k,l = inds
            if l < i:
                return (l,k,j,i), (3,2,1,0)
        elif t == 'y':
            i,j,k,l = inds
            if l < k:
                return (i,j,l,k), (0,1,3,2)
        return inds, tuple(range(len(inds)))
    @classmethod
    def _simple_presorts(cls, coord_types, inds1, inds2):
        # prealign coords
        t1, t2 = coord_types
        if t1 == 'r':
            i,j = inds1
            if t2 == 'r':
                ip, jp = inds2
                if jp == i:
                    inds2 = jp,ip
            elif t2 == 'a':
                ip,jp,kp = inds2
                if i == kp:
                    inds2 = kp,jp,ip
        elif t1 == 'a':
            i,j,k = inds1
            if t2 == 't':
                ip,jp,kp,lp = inds2
                if (jp == j and kp == i) or (kp == j and lp == i):
                    inds1 = k,j,i
        return inds1, inds2

    @classmethod
    def _get_coord_key(cls, inds1:'Iterable[int]', inds2:'Iterable[int]', coord_types=None):
        # we infer which type of element to use and return that
        # with appropriate substitutions

        if coord_types is None:
            coord_types = [None, None]
        coord_types = list(coord_types)
        if coord_types[0] is None:
            coord_types[0] = cls.infer_coord_type(inds1)
        if coord_types[1] is None:
            coord_types[1] = cls.infer_coord_type(inds2)
        type1, type2 = coord_types
        sorting = ['r', 'a', 't', 'y']
        swapped = sorting.index(type1) > sorting.index(type2)
        if swapped:
            inds1, inds2 = inds2, inds1
            type1, type2 = type2, type1

        inds1, inds2 = cls._simple_presorts((type1, type2), inds1, inds2)
        intersection = np.intersect1d(inds1, inds2)

        remapping = {}
        n_inter = 1
        n_diff = len(intersection) + 1
        i1, n_inter, n_diff = cls._remap_inter_inds(inds1, intersection, remapping, n_inter, n_diff)
        i2, n_inter, n_diff = cls._remap_inter_inds(inds2, intersection, remapping, n_inter, n_diff)

        i1, p1 = cls._canonicalize_inds(type1, i1)
        i2, p2 = cls._canonicalize_inds(type2, i2)

        return ((type1, type2), i1, i2), remapping, swapped, (p1, p2)

    @classmethod
    def kinetic_exprs(cls, inds1: 'Iterable[int]', inds2: 'Iterable[int]', coord_types=None, target_symbols=None):
        key, mapping, swapped, _ = cls._get_coord_key(inds1, inds2, coord_types=coord_types)
        (expr_g, expr_vp), perm = KEData.find_expressions(key, return_permutation=True)
        if perm is not None:
            # shared_indices = np.intersect1d(inds1, inds2)
            p1, p2 = perm
            if swapped:
                p1, p2 = p2, p1
            if p1 is not None:
                inds1 = [inds1[p] for p in p1]
            if p2 is not None:
                inds2 = [inds2[p] for p in p2]
            key, mapping, swapped, _ = cls._get_coord_key(inds1, inds2, coord_types=coord_types)
            # if p1 is not None:
            #     updates = {}
            #     for i, p in enumerate(p1):
            #         old = inds1[i]
            #         new = inds1[p]
            #         if (
            #                 old not in shared_indices and new not in shared_indices
            #                 or old in shared_indices and new in shared_indices
            #         ):
            #             updates[new] = mapping[old]
            #     mapping.update(updates)
            # if p2 is not None:
            #     updates = {}
            #     for i, p in enumerate(p2):
            #         old = inds2[i]
            #         new = inds2[p]
            #         if (
            #                 old not in shared_indices and new not in shared_indices
            #                 or old in shared_indices and new in shared_indices
            #         ):
            #             updates[new] = mapping[old]
            #     mapping.update(updates)
            # print(perm)
            # print(mapping)
            # print(expr)

        if not nput.is_numeric(expr_g):
            expr_g = cls._apply_remapping(expr_g, mapping, target_symbols)
        if not nput.is_numeric(expr_vp):
            expr_vp = cls._apply_remapping(expr_vp, mapping, target_symbols)

        return expr_g, expr_vp

    @classmethod
    def _apply_remapping(cls, expr, mapping, target_symbols):
        rev_mapping = {v: k for k, v in mapping.items()}
        subs = tuple(
            (s, cls.reindex_symbol(s, rev_mapping, target_symbols=target_symbols)) for s in expr.free_symbols)
        expr = expr.subs([(s, q) for s, (q, _) in subs])
        expr = expr.subs([(q, f) for s, (q, f) in subs])
        return expr
    @classmethod
    def _g_expr_direct(cls, coord_types, inds1: 'Iterable[int]', inds2: 'Iterable[int]'):
        dvs1 = InternalJacobianDisplacements.displacement_vectors(inds1, coord_type=coord_types[0])
        dvs2 = InternalJacobianDisplacements.displacement_vectors(inds2, coord_type=coord_types[1])
        total_expr = 0
        for n,x in enumerate(inds1):
            try:
                m = inds2.index(x)
            except ValueError:
                pass
            else:
                from .AnalyticJacobianDotCalculator import BondVector
                with BondVector.logger.block():
                    BondVector.logger.log_print("{woof}", woof=dvs1[n])
                    BondVector.logger.log_print("{woof}", woof=dvs2[m])
                    subcontrib = dvs1[n].dot(dvs2[m])
                    if hasattr(subcontrib, 'terms'):
                        subcontrib = sum(subcontrib.terms)
                    BondVector.logger.log_print("{subcontrib}", subcontrib=subcontrib)
                # subcontrib = dvs1[n].dot(dvs2[m])
                # if hasattr(subcontrib, 'simplify'):
                #     subcontrib = subcontrib.simplify()
                total_expr += subcontrib / cls.symbolic_m(x)
        return total_expr

    @classmethod
    def _vp_expr_direct(cls,  coord_types:'[str,str]', inds1: 'Iterable[int]', inds2: 'Iterable[int]', G):
        if G == 0:
            return 0
        t1, t2 = coord_types
        sym1 = cls.var(*inds1, coord_type=t1)
        sym2 = cls.var(*inds2, coord_type=t2)
        t1, inds1 = AnalyticModel.parse_symbol(sym1)
        t2, inds2 = AnalyticModel.parse_symbol(sym2)
        if t1 == "r":
            if t2 == "r":
                if inds1 == inds2:
                    return 0
                else:
                    return 1/(sym1*sym2) * G
            elif t2 == "a":
                return 1/(2*sym1) *(cls.sym.cot(sym2)*G + cls.sym.diff(G, sym2))
            elif t2 == "t":
                return 1/(2*sym1) *(cls.sym.diff(G, sym2))
            elif t2 == "y":
                return 1/(2*sym1) *(cls.sym.diff(G, sym2))
            else:
                raise ValueError("don't understand coord type '{}'".format(coord_types[1]))
        elif t1 == "a":
            if t2 == "a":
                if inds1 == inds2:
                    return (cls.sym.cot(sym1)*cls.sym.diff(G, sym1) - ((cls.sym.cot(sym1)**2)/2 + 1)) / 4
                else:
                    return (cls.sym.cot(sym1)*cls.sym.cot(sym2)*G
                            + cls.sym.cot(sym1)*cls.sym.diff(G, sym2)
                            + cls.sym.cot(sym2)*cls.sym.diff(G, sym1)) / 4
            elif t2 == "t":
                return cls.sym.cot(sym1)*cls.sym.diff(G, sym2)
            elif t2 == "y":
                return cls.sym.cot(sym1)*cls.sym.diff(G, sym2)
            else:
                raise ValueError("don't understand coord type '{}'".format(coord_types[1]))
        elif t1 == "t":
            return 0
        elif t1 == "y":
            return 0
        else:
            raise ValueError("don't understand coord type '{}'".format(coord_types[1]))
    @classmethod
    def kinetic_exprs_direct(cls, inds1: 'Iterable[int]', inds2: 'Iterable[int]',
                             coord_types=None, return_vp=True,
                             target_symbols=None
                             ):
        """
        Evaluated using the simple expressions in Table 1 from Frederick and Woywood
        :param inds1:
        :param inds2:
        :param coord_types:
        :return:
        """
        (coord_types, inds1, inds2), mapping, swapped, _ = cls._get_coord_key(inds1, inds2, coord_types=coord_types)
        # raise Exception(key, swapped)
        # if coord_types is None:
        #     coord_types = [None, None]
        #     if coord_types[0] is None:
        #         coord_types[0] = cls.infer_coord_type(inds1)
        #     if coord_types[1] is None:
        #         coord_types[1] = cls.infer_coord_type(inds2)

        # type1, type2 = coord_types
        # sorting = ['r', 'a', 't', 'y']
        # swapped = sorting.index(coord_types[0]) > sorting.index(coord_types[1])
        # if swapped:
        #     inds1, inds2 = inds2, inds1
        #     type1, type2 = type2, type1
        # raise Exception(key)
        G = cls._g_expr_direct(coord_types, inds1, inds2)
        if return_vp:
            vp = cls._vp_expr_direct(coord_types, inds1, inds2, G)
        else:
            vp = None

        # raise Exception(G, mapping)

        if not nput.is_numeric(G):
            G = cls._apply_remapping(G, mapping, target_symbols)
        if vp is not None and not nput.is_numeric(vp):
            vp = cls._apply_remapping(vp, mapping, target_symbols)

        return G, vp
    @classmethod
    def g(cls, inds1:'Iterable[int]', inds2:'Iterable[int]', coord_types=None, target_symbols=None, return_function=False,
          method='lookup'):
        if method == 'direct':
            expr = cls.kinetic_exprs_direct(inds1, inds2, coord_types=coord_types, return_vp=False)[0]
        else:
            expr = cls.kinetic_exprs(inds1, inds2, coord_types=coord_types, target_symbols=target_symbols)[0]
        if return_function:
            expr = GeometricFunction.from_expr(expr)
        return expr
    @classmethod
    def vp(cls, inds1:'Iterable[int]', inds2:'Iterable[int]', coord_types=None, target_symbols=None, return_function=False,
           method='lookup'):
        if method == 'direct':
            expr = cls.kinetic_exprs_direct(inds1, inds2, coord_types=coord_types, return_vp=True)[1]
        else:
            expr = cls.kinetic_exprs(inds1, inds2, coord_types=coord_types, target_symbols=target_symbols)[1]
        if return_function:
            expr = GeometricFunction.from_expr(expr)
        return expr

    @classmethod
    def g_matrix(cls, coord_specs, return_function=False, return_matrix=True, method='lookup'):
        ndim = len(coord_specs)
        rows, cols = np.triu_indices(ndim)
        exprs = [cls.g(coord_specs[i], coord_specs[j], method=method) for i,j in zip(rows, cols)]
        if return_matrix:
            mat = [[0]*ndim for _ in range(ndim)]
            for i,j,e in zip(rows, cols, exprs):
                mat[i][j] = mat[j][i] = e
            exprs = mat
        exprs = AnalyticModelBase.sym.Array(exprs)
        if return_function:
            exprs = GeometricFunction.from_expr(exprs)
        return exprs

    @classmethod
    def _g(cls, inds1:'Iterable[int]', inds2:'Iterable[int]'):
        inds1 = list(inds1)
        inds2 = list(inds2)
        if len(inds1) < len(inds2):
            iter_inds = inds1
            rem_inds = inds2
        else:
            iter_inds = inds2
            rem_inds = inds1
        d1 = cls.displacement_vectors(iter_inds)
        d2 = cls.displacement_vectors(rem_inds)
        g_contribs = []
        for n,i in enumerate(iter_inds):
            try:
                m = rem_inds.index(i)
            except ValueError:
                pass
            else:
                # raise Exception(d1[n], d2[m])
                a = d1[n]
                b = d2[m]
                # print(a, b)
                g_contribs.append(1/cls.symbolic_m(i)*a.dot(b))

        g = sum(g_contribs)
        if isinstance(g, AnalyticModelBase.sym.Expr):
            g = g.expand().simplify().expand()#.subs(sym.Abs, sym.Id).simplify()
        return g

class AnalyticModel:
    """
    Provides a symbolic representation of an analytically evaluatable Hamiltonian
    which can be used to get derived expressions to evaluate.
    Supplies methods to automatically run DVR and VPT calculations from the model
    specifications as well.

    :related:AnalyticPotentialConstructor, AnalyticPotentialConstructor
    """

    def __init__(self, coordinates, potential, dipole=None, values=None, rotation=None):
        self.coords = coordinates
        self.vals = values
        self._syms = None
        self._base_g = None
        self._g = None
        self._u = None
        self.pot = potential
        self.dip = dipole
        if rotation is not None:
            if len(rotation) == 2:
                r, i = rotation
                r = np.asanyarray(r)
                i = np.asanyarray(i)
                if r.ndim == 2:
                    rotation = r
                    inverse = i
                else:
                    inverse = np.linalg.inv(rotation)
            else:
                inverse = np.linalg.inv(rotation)
        else:
            inverse = None
        self.rotation = rotation
        self.inverse = inverse
        self._rot_coords = None

    @classmethod
    def from_potential(cls, potential, dipole=None, values=None, rotation=None):
        # we infer the coordinates from the terms that appear in the potential

        coords = [
            x for x in
            potential.free_symbols
            if x.name.split("[", 1)[0] in {"r", "a", "t", "y"}
        ]
        return cls(
            coords,
            potential,
            dipole=dipole,
            values=values,
            rotation=rotation
        )


    @property
    def internal_coordinates(self):
        if self._syms is None:
            self._load_symbols()
        return self._syms[0]
    @property
    def constants(self):
        return {k:self.vals[k] for k in self.vals.keys() - set(self.internal_coordinates)}

    def normal_modes(self, dimensionless=True):
        from ..Modes import NormalModes

        v = self.v(evaluate=True)[-1]
        g = self.g(evaluate=True)[-1]

        freqs, modes, mode_inv = NormalModes.get_normal_modes(v, g, dimensionless=dimensionless)
        # modes = modes.T
        # mode_inv = mode_inv.T
        # freqs, modes,
        #
        # freqs2, mode_inv = scipy.linalg.eigh(v, g, type=2) # modes come out as a (internals, normal_mode) array
        # freqs = np.sqrt(freqs2)
        #
        # # normalization = np.broadcast_to(1 / np.linalg.norm(modes, axis=0), modes.shape)
        # # mode_inv = mode_inv  # now as (normal_mode, internals) with frequency dimension removed
        # modes = np.linalg.inv(mode_inv)
        # if dimensionless:
        #     mode_inv = mode_inv * np.sqrt(freqs[np.newaxis, :])
        #     modes = modes / np.sqrt(freqs[:, np.newaxis])
        # else:
        #     modes = modes / np.linalg.norm(modes, axis=1)[:, np.newaxis]
        #     mode_inv = mode_inv / np.linalg.norm(mode_inv, axis=0)[np.newaxis, :]

        #
        #
        # # coords = [AnalyticModelBase.dot(v, self.coords) for v in modes]

        return freqs, modes, mode_inv

    def to_normal_modes(self, dimensionless=True):
        import copy

        if self.rotation is not None:
            raise ValueError("already rotated")
        else:
            freqs, modes, mode_inv = self.normal_modes(dimensionless=dimensionless)
            new = copy.copy(self)
            new._base_g = new._g = new._u = None
            new.rotation = modes
            new.inverse = mode_inv
            return new, freqs

    def get_VPT_expansions(self,
                           order=2,
                           expansion_order=None,
                           include_potential=None,
                           include_gmatrix=None,
                           include_pseudopotential=None,
                           evaluate=True
                           ):
        if self.rotation is None:
            self, _ = self.to_normal_modes()

        if expansion_order is None:
            expansion_order = {}
        if 'potential' in expansion_order:
            pot_order = expansion_order['potential']
        elif 'default' in expansion_order:
            pot_order = expansion_order['default']
        else:
            pot_order = order
        if 'kinetic' in expansion_order:
            ke_order = expansion_order['kinetic']
        elif 'default' in expansion_order:
            ke_order = expansion_order['default']
        else:
            ke_order = order
        if 'pseudopotential' in expansion_order:
            u_order = expansion_order['pseudopotential']
        elif 'default' in expansion_order:
            u_order = expansion_order['default']
        else:
            u_order = order

        terms = {}
        if include_gmatrix is not False:
            terms['kinetic_terms'] = self.g(ke_order, evaluate=evaluate)
        if include_potential is not False:
            terms['potential_terms'] = self.v(pot_order+2, evaluate=evaluate)[2:]
        if include_pseudopotential is not False:
            terms['pseudopotential_terms'] = self.vp(u_order, evaluate=evaluate)


        if 'dipole' in expansion_order:
            d_order = expansion_order['dipole']
        elif 'default' in expansion_order:
            d_order = expansion_order['default']
        else:
            d_order = order
        if self.dip is not None:
            terms['dipole_terms'] = self.mu(d_order+1, evaluate=evaluate)
        return terms

    def run_VPT(self,
                order=2,
                states=2,
                return_analyzer=True,
                expansion_order=None,
                include_potential=None,
                include_gmatrix=None,
                include_pseudopotential=None,
                atoms=None, coords=None,
                **kwargs
    ):
        from ..VPT2 import VPTAnalyzer, VPTRunner

        if self.rotation is None:
            self, _ = self.to_normal_modes()
        expansions = self.get_VPT_expansions(order=order,
                                             expansion_order=expansion_order,
                                             include_potential=include_potential,
                                             include_gmatrix=include_gmatrix,
                                             include_pseudopotential=include_pseudopotential,
                                             evaluate=True
                                             )
        freqs = np.diag(expansions['potential_terms'][0])
        ncoords = len(freqs)
        natoms = (ncoords + 6) // 3
        if return_analyzer:
            runner = VPTAnalyzer.run_VPT
        else:
            runner = VPTRunner.run_simple
        if atoms is None:
            atoms = ["H"] * natoms
        if coords is None:
            coords = np.zeros((natoms, 3))
        analyzer = runner(
            [atoms, coords],  # dummy data since this won't be used at all
            states,
            modes={
                "freqs": freqs,
                "matrix": np.zeros((3 * natoms, len(freqs)))  # dummy data since this won't be used at all
            },
            order=order,
            **dict(expansions, **kwargs),
            expansion_order=expansion_order,
            include_potential=include_potential,
            include_gmatrix=include_gmatrix,
            include_pseudopotential=include_pseudopotential,
            include_coriolis_coupling=False,  # tell the job not to try to calculate the Coriolis coupling tensor
            calculate_intensities=self.dip is not None # no dipole so we can't get intensities
        )
        if return_analyzer:
            analyzer.print_output_tables(print_intensities=self.dip is not None, print_energies=self.dip is None)
        return analyzer

    def _get_rotated_coordinates(self):
        if self._rot_coords is None:
            self._rot_coords = AnalyticModelBase.transform_coordinates(AnalyticModelBase.transpose(self.rotation))
        return self._rot_coords
    def _get_inverse_coordinates(self):
        if self._inv_coords is None:
            self._inv_coords = AnalyticModelBase.transform_coordinates(self.inverse, coord_vec=self.coords)
        return self._inv_coords

    class SympyExpr:
        def __init__(self, expr, core, ndim):
            self.expr = expr
            self.lam = core
            self.ndim = ndim
        def _broadcast_tree(self, shape, expr_list):
            if not isinstance(expr_list, (np.ndarray,) + AnalyticModelBase.get_numeric_types()):
                return [self._broadcast_tree(shape, l) for l in expr_list]
            elif isinstance(expr_list, np.ndarray):
                return expr_list
            else:
                return np.full(shape, expr_list)

        def _array_call(self, grid, core):
            vals = core(*grid)
            # broadcast appropriately
            if not isinstance(vals, np.ndarray):
                if isinstance(vals, AnalyticModelBase.get_numeric_types()):
                    vals = np.full(grid.shape[1:], vals)
                else:
                    vals = np.array(self._broadcast_tree(grid.shape[1:], vals))
                    for _ in range(grid.ndim - 1):
                        vals = np.moveaxis(vals, -1, 0)
            return vals
        def _rec_call(self, grid, core, depth=0, rec_counter=None):
            if callable(core):
                if rec_counter is not None:
                    rec_counter[0] = depth
                return self._array_call(grid, core)
            else:
                if rec_counter is None:
                    rec_counter = [-1]
                return [self._rec_call(grid, c, depth=depth+1, rec_counter=rec_counter) for c in core]
        def __call__(self, grid, vector=None, **kwargs):
            core = self.lam
            ndim = self.ndim
            if hasattr(grid, 'ndim'):
                ndims = [grid.ndim]
            else:
                grid = [np.asanyarray(g) for g in grid]
                ndims = np.unique([c.ndim for c in grid])
            if len(ndims) > 1: # inhomogenous
                vals = core(*grid)
                if not nput.is_numeric(vals):
                    vals = np.asanyarray(vals)
                return vals
            else:
                grid = np.asanyarray(grid)
                if grid.ndim == 1:
                    vals = core(*grid)
                    if not isinstance(vals, AnalyticModelBase.get_numeric_types()):
                        vals = np.asanyarray(vals)
                    return vals

                if vector is None:
                    vector = grid.shape[-1] == ndim and grid.shape[0] != ndim
                if vector:
                    grid = np.moveaxis(grid, -1, 0) #grid.transpose(np.roll(np.arange(grid.ndim), 1))

                rec_counter = [0] # ugly hack to track the number of recursions done over lists
                base = self._rec_call(grid, core, rec_counter=rec_counter)
                res = np.asanyarray(base)
                if not isinstance(base, np.ndarray):
                    # raise Exception(
                    #     res.shape,
                    #
                    # )
                    # print(
                    #     np.roll(np.arange(res.ndim), -rec_counter[0]),
                    #     rec_counter[0]
                    # )
                    res = res.transpose(
                        np.roll(np.arange(res.ndim), -rec_counter[0])
                    )

            return res

        def __repr__(self):
            return "{}({})".format(
                type(self).__name__,
                self.expr
            )

        @classmethod
        def _lambdiy(cls, vars, expr, lambdify_arrays=False):
            types = (
                (AnalyticModelBase.sym.Expr, AnalyticModelBase.sym.Array)
                    if lambdify_arrays else
                (AnalyticModelBase.sym.Expr,)
            )
            if isinstance(expr,types):
                core = AnalyticModelBase.sym.lambdify(vars, expr)
                return core
            else:
                return [cls._lambdiy(vars, e) for e in expr]
        @classmethod
        def lambdify(cls, vars, expr, ndim, lambdify_arrays=False):
            core = cls._lambdiy(vars, expr, lambdify_arrays=lambdify_arrays)
            return cls(expr, core, ndim)

    @classmethod
    def prep_lambda_expr(cls, base_coords, expr, dummify=False, rewrite_trig=True):
        sympy = AnalyticModelBase.sym
        if dummify:
            vars = sympy.symbols([f"{k}{i}" for i, k in enumerate([str(crd)[0] for crd in base_coords])])
            expr = expr.subs(
                zip(base_coords, vars)
            )
            base_coords = vars
        if rewrite_trig:
            expr = expr.rewrite(
                sympy.cot, sympy.tan
            ).rewrite(
                sympy.csc, sympy.sin
            ).rewrite(
                sympy.sec, sympy.cos
            )
        return base_coords, expr
    @classmethod
    def lambdify(cls, coord_vec, expr, coordinate_transform=None, mode=None,
                 dummify=False,
                 rewrite_trig=True,
                 lambdify_arrays=False
                 ):
        #TODO: make consistent with wrap_function
        coord_vec, expr = cls.prep_lambda_expr(coord_vec, expr, dummify=dummify, rewrite_trig=rewrite_trig)
        ndim = len(coord_vec)
        if isinstance(expr, AnalyticModelBase.get_numeric_types()):
            if isinstance(expr, np.ndarray):
                val = expr
            else:
                val = float(expr)
            return cls.SympyExpr(val, lambda c, *r: np.full(c.shape, val), ndim)
        if coordinate_transform is not None:
            raise NotImplementedError('generic coordinate transforms not supported yet...')
            coord_vec, new_coords = coordinate_transform
            # coord_vec, new_coords = self._get_rotated_coordinates()
            # subtract off equilibrium values since we have a rotation...
            shift_subs = [(s, s + self.vals[s]) for s in self.coords if s in self.vals]
            expr = expr.subs(shift_subs)
            expr = expr.subs([(old, new) for old, new in zip(self.coords, new_coords)])

        return cls.SympyExpr.lambdify(coord_vec, expr, ndim, lambdify_arrays=lambdify_arrays)

    def wrap_function(self, expr, transform_coordinates=True, mode=None):
        coord_vec = self.coords
        ndim = len(coord_vec)
        if isinstance(expr, AnalyticModelBase.get_numeric_types()):
            if isinstance(expr, np.ndarray):
                val = expr
            else:
                val = float(expr)
            return self.SympyExpr(val, lambda c,*r:np.full(c.shape, val), ndim)

        if transform_coordinates and self.rotation is not None:
            coord_vec, new_coords = self._get_rotated_coordinates()
            # subtract off equilibrium values since we have a rotation...
            shift_subs = [(s, s+self.vals[s]) for s in self.coords if s in self.vals]
            expr = expr.subs(shift_subs)
            expr = expr.subs([(old, new) for old,new in zip(self.coords, new_coords)])

        return self.SympyExpr.lambdify(coord_vec, expr, ndim)

    def expand_potential(self, order, lambdify=True, evaluate=True, contract=True):
        potential_expansions = self.v(order, evaluate=evaluate)
        coord_vec = self.coords
        if contract:
            pots = []
            for i, d in enumerate(potential_expansions):
                for j in range(i):
                    d = AnalyticModelBase.dot(coord_vec, d, axes=[0, -1])
                pots.append(1/math.factorial(i)*d)
        else:
            pots = potential_expansions
        if lambdify:
            pots = self.wrap_function(sum(pots)) if contract else [self.wrap_function(p) for p in pots]

        return pots
    def get_DVR_parameters(self,
                           expansion_order=None,
                           lambdify=True,
                           evaluate='constants'
                           ):
        if expansion_order is not None:
            potential_function = self.expand_potential(expansion_order, lambdify=lambdify)
        else:
            if evaluate:
                potential_function = self.evaluate(self.pot, mode=evaluate)
            else:
                potential_function = self.pot
            if lambdify:
                potential_function = self.wrap_function(potential_function)
        gds = self.g(order=2, evaluate=evaluate)
        g = gds[0]
        g_deriv = [gds[2][i][i][i][i] for i in range(len(self.coords))]

        # print(g)
        # print(g_deriv)
        if lambdify:
            g = [
                [
                    self.wrap_function(g[i][j])
                    for j in range(len(self.coords))
                ]
                for i in range(len(self.coords))
            ]
            g_deriv = [
                self.wrap_function(g_deriv[i])
                for i in range(len(self.coords))
            ]
        # if len(g_deriv) == 1:
        #     g = g[0][0]
        #     g_deriv = g_deriv[0]
        return {
            'potential_function':potential_function,
            'g':g,
            'g_deriv':g_deriv
        }
        # raise Exception(potential_function, g)
    def setup_DVR(self,
                  domain=None, divs=None,
                  use_normal_modes=False,
                  expansion_order=None,
                  potential_function=None,
                  **params
                  ):
        if use_normal_modes:
            if self.rotation is None:
                self, _ = self.to_normal_modes(dimensionless=True)
            return self.setup_DVR(
                domain=domain, divs=divs,
                use_normal_modes=False,
                expansion_order=expansion_order,
                **params
            )

        from ..DVR import DVR
        params = dict(self.get_DVR_parameters(expansion_order=expansion_order), **params)
        if potential_function is not None:
            params['potential_function'] = potential_function
        return DVR(
            domain=domain,
            divs=divs,
            **params
        )

    def evaluate(self, expr, mode='all', numericize=False):
        if self.vals is None:
            raise ValueError("can't evaluate without values for parameters")
        if mode == True:
            mode = 'all'
        a = AnalyticModelBase.eval_exprs(expr, self.constants if mode == 'constants' else self.vals)
        if mode != 'constants':
            try:
                a = np.array(a, dtype=float)
            except ValueError:
                pass
            else:
                if a.ndim == 0:
                    a = a.tolist()
        return a

    def _load_symbols(self):
        sym_list = []
        sym_set = set()
        for x in self.coords:
            for s in x.free_symbols:
                if s not in sym_set:
                    sym_set.add(s)
                    sym_list.append(s)
        self._syms = (tuple(sym_list), tuple(self.parse_symbol(s) for s in sym_list))

    @classmethod
    def parse_symbol(self, sym):
        name = sym.name
        t, a = name.split("[")
        a = a.strip("]").split(",")
        return (t, tuple(int(x) for x in a))

    def jacobian(self, order=0, evaluate=False, lambdify=False):
        ics = self.internal_coordinates
        crd = self.coords
        jac = AnalyticModelBase.sym.Matrix([AnalyticModelBase.take_derivs(c, ics) for c in crd]).transpose()
        if order > 0:
            if self.rotation is not None:
                jac = AnalyticModelBase.dot(jac, self.rotation)
            for i in range(order):
                jac = AnalyticModelBase.take_derivs(jac, ics)
        elif evaluate and self.rotation is not None:
                jac = AnalyticModelBase.dot(self.evaluate(jac, mode=evaluate), self.rotation)
        if evaluate:
            jac = self.evaluate(jac, mode=evaluate)
        else:
            jac = AnalyticModelBase.sym.Matrix(jac)
        if lambdify:
            jac = self.wrap_function(jac)
        return jac
    def jacobian_inverse(self, order=0, evaluate=False):
        ics = self.internal_coordinates
        crd = self.coords
        # try:
        #     jac = sym.Array([AnalyticModelBase.take_derivs(c, crd) for c in ics])
        #     for i in range(order):
        #         jac = AnalyticModelBase.take_derivs(jac, crd)
        # except ValueError:
        jac = self.jacobian().inv()
        if order > 0:
            if self.inverse is not None:
                jac = AnalyticModelBase.dot(self.inverse, jac)
            for i in range(order):
                jac = AnalyticModelBase.take_derivs(jac, crd)
        elif evaluate and self.inverse is not None:

                jac = AnalyticModelBase.dot(self.inverse, self.evaluate(jac, mode=evaluate))
        if evaluate:
            jac = self.evaluate(jac, mode=evaluate)
        else:
            jac = AnalyticModelBase.sym.Matrix(jac)
        # if self.inverse is not None:
        #     jac = AnalyticModelBase.dot(jac, self.inverse)
        # for i in range(order):
        #     jac = AnalyticModelBase.take_derivs(jac, crd)
        # if evaluate:
        #     jac = self.evaluate(jac)
        # else:
        #     jac = sym.Matrix(jac)
        return jac
    def _base_gmat(self):
        if self._base_g is None:
            if self._syms is None:
                self._load_symbols()
            symlist = self._syms[1]
            targets = {s.name for s in self._syms[0]}
            if self.vals is not None:
                targets.update(s.name for s in self.vals.keys())
            return [[AnalyticKineticEnergyConstructor.g(a[1], b[1], coord_types=[a[0], b[0]], target_symbols=targets) for b in symlist] for a in symlist]
        return self._base_g
    def g(self, order=0, evaluate=False, lambdify=False):
        # Gmatrix elements will basically involve taking some
        # kind of direct product of coordinate reps
        J_t = self.jacobian_inverse(evaluate=evaluate)
        if self._g is None:
            G = self._base_gmat()
            J = J_t.transpose()
            self._g = AnalyticModelBase.dot(AnalyticModelBase.dot(J_t, G), J)
        all_gs = [self._g]
        if order > 0:
            J_inv = self.jacobian(evaluate=evaluate).transpose()
            for i in range(order):
                all_gs.append(AnalyticModelBase.take_derivs(all_gs[-1], self.internal_coordinates))
        if evaluate:
            for i,g in enumerate(all_gs):
                all_gs[i] = self.evaluate(g, mode=evaluate)
        for i in range(1, len(all_gs)):
            g = all_gs[i]
            for j in range(i):
                g = AnalyticModelBase.dot(J_inv, g, axes=[1, -3])
            all_gs[i] = g
        if lambdify:
            all_gs = [self.wrap_function(g) for g in all_gs]
        return all_gs
    def v(self, order=2, evaluate=False, lambdify=False):
        # we provide a Taylor series expansion of the potential
        v = self.pot
        all_vs = [v]
        if order > 0:
            J_inv = self.jacobian(evaluate=evaluate).transpose()
        for i in range(order):
            all_vs.append(AnalyticModelBase.take_derivs(all_vs[-1], self.internal_coordinates))
        if evaluate:
            for i,v in enumerate(all_vs):
                all_vs[i] = self.evaluate(v, mode=evaluate)
        for i in range(1, len(all_vs)):
            v = all_vs[i]
            for j in range(i):
                v = AnalyticModelBase.dot(J_inv, v, axes=[1, -1])
            all_vs[i] = v
        if lambdify:
            all_vs = [self.wrap_function(v) for v in all_vs]
        return all_vs

    def _base_u(self):
        if self._syms is None:
            self._load_symbols()
        symlist = self._syms[1]
        targets = {s.name for s in self._syms[0]}
        if self.vals is not None:
            targets.update(s.name for s in self.vals.keys())
        return [[AnalyticKineticEnergyConstructor.vp(a[1], b[1], coord_types=[a[0], b[0]], target_symbols=targets) for b in symlist] for a in symlist]
        # return [[AnalyticKineticEnergyConstructor.vp(a[1], b[1], coord_types=[a[0], b[0]]) for b in symlist] for a in symlist]
    def vp(self, order=0, evaluate=False, lambdify=False):
        J = self.jacobian()
        if self._u is None:
            U = self._base_u()
            J_t = AnalyticModelBase.transpose(J)
            self._u = 8*AnalyticModelBase.contract(AnalyticModelBase.dot(AnalyticModelBase.dot(J_t, U), J), (0, 1))
        u = self._u
        all_vs = [u]
        if order > 0:
            J_inv = self.jacobian(evaluate=evaluate).transpose()
        for i in range(order):
            all_vs.append(AnalyticModelBase.take_derivs(all_vs[-1], self.internal_coordinates))
        if evaluate:
            for i, v in enumerate(all_vs):
                all_vs[i] = self.evaluate(v, mode=evaluate)
        for i in range(1, len(all_vs)):
            v = all_vs[i]
            for j in range(i):
                v = AnalyticModelBase.dot(J_inv, v, axes=[1, -1])
            all_vs[i] = v
        if lambdify:
            all_vs = [self.wrap_function(u) for u in all_vs]
        return all_vs

    def mu(self, order=1, evaluate=False, lambdify=False):
        # we provide a Taylor series expansion of the dipole
        mu = []
        for v in self.dip:
            all_vs = [v]
            if order > 0:
                J_inv = self.jacobian(evaluate=evaluate).transpose()
            for i in range(order):
                all_vs.append(AnalyticModelBase.take_derivs(all_vs[-1], self.internal_coordinates))
            if evaluate:
                for i, v in enumerate(all_vs):
                    all_vs[i] = self.evaluate(v, mode=evaluate)
            for i in range(1, len(all_vs)):
                v = all_vs[i]
                for j in range(i):
                    v = AnalyticModelBase.dot(J_inv, v, axes=[1, -1])
                all_vs[i] = v
            if lambdify:
                all_vs = [self.wrap_function(u) for u in all_vs]
            mu.append(all_vs)
        return mu

    Potential = AnalyticPotentialConstructor
    morse = AnalyticPotentialConstructor.morse
    harmonic = AnalyticPotentialConstructor.harmonic
    linear = AnalyticPotentialConstructor.linear
    power = AnalyticPotentialConstructor.power
    sin = AnalyticPotentialConstructor.sin
    cos = AnalyticPotentialConstructor.cos
    KE = AnalyticKineticEnergyConstructor

    class NamespaceContext:
        def __init__(self, context=None):
            self._context = context
            self._additions = set()
        def _get_frame_vars(self):
            import inspect
            frame = inspect.currentframe()
            parent = frame.f_back.f_back.f_back
            # print(parent.f_locals)
            return parent.f_locals
        def insert_vars(self):
            globs = self._context
            if globs is None:
                globs = self._get_frame_vars()
            for k, v in dict(
                    r=AnalyticModel.r,
                    a=AnalyticModel.a,
                    cos=AnalyticModel.cos,
                    sin=AnalyticModel.sin,
                    morse=AnalyticModel.morse,
                    harmonic=AnalyticModel.harmonic,
                    sym=AnalyticModel.sym,
                    m=AnalyticModel.m,
            ).items():
                globs[k] = v
                self._additions.add(k)
        def prune_vars(self):
            globs = self._context
            if globs is None:
                globs = self._get_frame_vars()
            for x in self._additions:
                try:
                    del globs[x]
                except KeyError:
                    pass
        def __enter__(self):
            self.insert_vars()
        def __exit__(self, exc_type, exc_val, exc_tb):
            self.prune_vars()

    @classmethod
    def sym(self, base, *args):
        return AnalyticModelBase.symbol(base, *args)
    @classmethod
    def m(self, i):
        return AnalyticModelBase.symbolic_m(i)
    @classmethod
    def r(self, i, j):
        return AnalyticModelBase.symbolic_r(i, j)
    @classmethod
    def a(self, i, j, k):
        return AnalyticModelBase.symbolic_a(i, j, k)
    @classmethod
    def t(self, i, j, k, l):
        return AnalyticModelBase.symbolic_t(i, j, k, l)
    @classmethod
    def y(self, i, j, k, l):
        return AnalyticModelBase.symbolic_t(i, j, k, l)

    convert = UnitsData.convert
    @staticmethod
    def mass(atom):
        return AtomData[atom]["Mass"] * UnitsData.convert("AtomicMassUnits", "AtomicUnitOfMass")

    def molecular_potential(self, mol):
        if mol.internals is None:
            raise ValueError("Molecule {} needs internal coordinates to be used with a model potential")
        return MolecularModelPotentialFunction(self, mol)
    def molecular_dipole(self, mol):
        if mol.internals is None:
            raise ValueError("Molecule {} needs internal coordinates to be used with a model dipole")
        return MolecularModelDipoleFunction(self, mol)
    def molecular_gmatrix(self, mol):
        if mol.internals is None:
            raise ValueError("Molecule {} needs internal coordinates to be used with a model gmatrix")
        return MolecularModelGMatrixFunction(self, mol)

class MolecularModel(AnalyticModel):
    def __init__(self, mol, coords, potential, dipole=None, values=None, rotation=None):
        super().__init__(coords, potential, dipole=dipole, values=values, rotation=rotation)
        self.mol = mol
    @property
    def potential(self):
        return MolecularModelPotentialFunction(self, self.mol)
    @property
    def gmatrix(self):
        return MolecularModelGMatrixFunction(self, self.mol)
    @property
    def vprime(self):
        return MolecularModelVPrimeFunction(self, self.mol)
    @property
    def dipole(self):
        return MolecularModelDipoleFunction(self, self.mol)

    def setup_AIMD(
            self,
            timestep=.5,
            initial_energies=None,
            initial_displacements=None,
            displaced_coords=None,
            track_kinetic_energy=False,
            track_velocities=False
    ):

        mol = self.mol
        pot_func = self.potential

        return mol.setup_AIMD(
            pot_func,
            timestep=timestep,
            initial_energies=initial_energies,
            initial_displacements=initial_displacements,
            displaced_coords=displaced_coords,
            track_kinetic_energy=track_kinetic_energy,
            track_velocities=track_velocities
        )

    def setup_DGB(
            self,
            centers,
            *,
            masses=None,
            modes='normal',
            transformations=None,
            alphas='auto',
            cartesians=None,
            potential_function=None,
            dipole_function=None,
            optimize_centers=None,
            quadrature_degree=None,
            expansion_degree=None,
            pairwise_potential_functions=None,
            internals=False,
            logger=True,
            **opts
    ):

        from ..DGB import DGB

        if internals is True:
            raise NotImplementedError("internal coordinate DGB support coming soon? hopefully?")

        defaults = dict(
            alphas=alphas,
            modes=modes,
            transformations=transformations,
            cartesians=cartesians,
            logger=logger,
            optimize_centers=optimize_centers,
            quadrature_degree=quadrature_degree,
            expansion_degree=expansion_degree,
            pairwise_potential_functions=pairwise_potential_functions
        )
        defaults = {k: v for k, v in defaults.items() if v is not None}
        opts = dict(defaults, **opts)

        if masses is None:
            masses = self.mol.atomic_masses

        if potential_function is None:
            potential_function = self.potential
        if dipole_function is None:
            dipole_function = lambda coords, deriv_order=None: (
                self.dipole(coords)
                    if deriv_order is None else
                [np.moveaxis(d, -1, 1) for d in self.dipole(coords, deriv_order=deriv_order)]
        )

        return DGB.construct(
            centers,
            potential_function,
            masses=masses,
            dipole_function=dipole_function,
            **opts
        )

class MolecularModelFunction:

    def __init__(self, deriv_evaluator, mol):
        self.mol = mol
        self.deriv_evaluator = deriv_evaluator
        self.derivs = []

    def evaluate(self, carts, deriv_order=None, internals=False, which=None, sel=None, axes=None, derivs=None):
        carts = np.asanyarray(carts)

        ads = 1 if deriv_order is None else deriv_order + 1
        if derivs is None:
            if len(self.derivs) < ads + 1:
                self.derivs = self.deriv_evaluator(order=ads-1, evaluate='constants', lambdify=True)
            derivs = self.derivs
        elif len(derivs) < ads:
            raise ValueError("need at least {} derivs".format(ads - 1))


        base_shape = carts.shape[:-2]
        carts = carts.reshape((-1,) + carts.shape[-2:])
        if (
                internals
                or which is not None
                or sel is not None
                or axes is not None
        ):
            carts = self.mol.get_displaced_coordinates(
                carts,
                which=which, sel=sel, axes=axes, use_internals=internals,
                shift=False
            )

        vals = self.mol.evaluate_at(
            lambda c, deriv_order=None: (
                derivs[0](c, vector=True)
                    if deriv_order is None else
                [
                    df(c, vector=True) for df in derivs[:deriv_order + 1]
                ]
            ),
            carts,
            deriv_order=None if deriv_order is not None and deriv_order == 0 else deriv_order,
            use_internals=True,
            strip_embedding=True
        )
        if deriv_order is not None:
            if deriv_order == 0:
                vals = [vals]

            take = None
            if which is not None:
                take = tuple(
                    np.ravel_multi_index(idx, (3, len(self.mol.masses)))
                    if not isinstance(idx, (int, np.integer)) else
                        idx
                    for idx in which
                )
            elif sel is not None or axes is not None:
                if sel is None:
                    sel = np.arange(len(self.mol.masses))
                if axes is None:
                    axes = np.arange(3)
                take = np.ravel_multi_index(np.array(list(itertools.product(sel, axes))).T, (3, len(self.mol.masses)))

            if take is not None:
                new = []
                for n, d in enumerate(vals):
                    for j in range(n):
                        d = np.take(d, take, axis=j + 1)
                    d = d.reshape(base_shape + d.shape[1:])
                    new.append(d)
                vals = new

            vals = [
                v.reshape(base_shape + v.shape[1:])
                for v in vals
            ]

        else:
            vals = vals.reshape(base_shape)

        return vals

    def __call__(self, carts, deriv_order=None, internals=False, which=None, sel=None, axes=None):
        return self.evaluate(carts, deriv_order=deriv_order, internals=internals, which=which, sel=sel, axes=axes)

class MolecularModelPotentialFunction(MolecularModelFunction):
    def __init__(self, model, mol):
        super().__init__(model.v, mol)

class MolecularModelDipoleFunction(MolecularModelFunction):
    def __init__(self, model, mol):
        super().__init__(model.mu, mol)
        self._subfuncs = []

    def evaluate(self, carts, deriv_order=None, internals=False, which=None, sel=None, axes=None):
        """
        This has the added complication of needing to dispatch over the axes...

        :param carts:
        :type carts:
        :param deriv_order:
        :type deriv_order:
        :param internals:
        :type internals:
        :param which:
        :type which:
        :param sel:
        :type sel:
        :param axes:
        :type axes:
        :return:
        :rtype:
        """
        ads = 1 if deriv_order is None else deriv_order + 1
        if len(self.derivs) < ads + 1:
            self.derivs = self.deriv_evaluator(order=ads - 1, evaluate='constants', lambdify=True)

        vals = []
        for d in self.derivs:
            vals.append(
                super().evaluate(carts, deriv_order=deriv_order, internals=internals, which=which, sel=sel, axes=axes, derivs=d)
            )

        if deriv_order is None:
            vals = np.moveaxis(np.array(vals), 0, -1)
        else:
            transpose_terms = [[] for _ in range(deriv_order + 1)]
            for term_list in vals:
                for i, t in enumerate(term_list):
                    transpose_terms[i].append(t)
            vals = [
                np.moveaxis(np.array(tl), 0, -1)
                for tl in transpose_terms
            ]

        return vals

class MolecularModelGMatrixFunction(MolecularModelFunction):
    def __init__(self, model, mol):
        super().__init__(model.g, mol)

class MolecularModelVPrimeFunction(MolecularModelFunction):
    def __init__(self, model, mol):
        super().__init__(model.vp, mol)